"""This module provides functions which determine various observatory
specific policies for JWST:

1. How to convert reference file basenames to fully specified paths.

2. How to manage parameters for reference file Validator objects used
in the certification of reference files.
"""
import os.path
import re
import warnings

from asdf.tags.core import NDArrayType
from stdatamodels.validate import ValidationWarning

# =======================================================================

from crds.core import rmap, config, utils, timestamp, log, exceptions
from crds.certify import generic_tpn
from crds import data_file
from crds.io import abstract

# =======================================================================

# These two functions decouple the generic reference file certifier program
# from observatory-unique ways of specifying and caching Validator parameters.

from crds.jwst import TYPES, INSTRUMENTS, FILEKINDS, EXTENSIONS, INSTRUMENT_FIXERS, TYPE_FIXERS

from . import schema

# definitively lists possible exp_types by instrument, or all instruments
from .schema import get_exptypes

get_row_keys_by_instrument = TYPES.get_row_keys_by_instrument
get_item = TYPES.get_item
suffix_to_filekind = TYPES.suffix_to_filekind
filekind_to_suffix = TYPES.filekind_to_suffix
get_all_tpninfos = TYPES.get_all_tpninfos

from crds.jwst.pipeline import header_to_reftypes, header_to_pipelines

# =======================================================================

MODEL = None

def get_datamodels():
    try:
        from jwst import datamodels  # this is fatal.
    except ImportError:
        log.error(
            "CRDS requires installation of the 'jwst' package to operate on JWST files.")
        raise
    global MODEL
    if MODEL is None:
        with log.error_on_exception(
                "Failed constructing basic JWST DataModel"):
            MODEL = datamodels.DataModel()
    return datamodels

# =============================================================================

HERE = os.path.dirname(__file__) or "."

def tpn_path(tpn_file):
    """Return the full filepath of `tpn_file`."""
    return os.path.join(HERE, "tpns", tpn_file)

def get_extra_tpninfos(refpath):
    """Returns TpnInfos (valid value enumerations) derived from the cal code data models schema."""
    return []
    # XXX Disabled during transition period to more specialized reference file schema
    # that will require enhanced schema scraping.
    # return schema.get_schema_tpninfos(refpath)

def project_check(refpath, rmap):
    if not rmap.filekind.startswith("pars-"):
        get_data_model_flat_dict(refpath)

def get_data_model_flat_dict(filepath):
    """Get the header from `filepath` using the jwst data model."""
    datamodels = get_datamodels()
    log.info("Checking JWST datamodels.")
    # with log.error_on_exception("JWST Data Model (jwst.datamodels)"):
    try:
        with datamodels.open(filepath) as d_model:
            flat_dict = d_model.to_flat_dict(include_arrays=False)
    except Exception as exc:
        raise exceptions.ValidationError("JWST Data Models:", str(exc).replace("u'","'")) from exc
    return flat_dict

# =======================================================================

def match_context_key(key):
    """Set the case of a context key appropriately for this project, JWST
    always uses upper case.
    """
    return key.upper()

# =======================================================================

REF_EXT_RE = re.compile(r"\.fits|\.r\dh$")

@utils.cached
def get_file_properties(filename):
    """Figure out (instrument, filekind, serial) based on `filename` which
    should be a mapping or FITS reference file.

    >> get_file_properties("./hst_acs_biasfile_0001.rmap")
    ('acs', 'biasfile')

    >> get_file_properties("./hst_acs_biasfile_0001.pmap")
    Traceback (most recent call last):
    ...
    AssertionError: Invalid .pmap filename './hst_acs_biasfile_0001.pmap'

    >> get_file_properties("test_data/s7g1700gl_dead.fits")
    """
    if config.is_mapping(filename):
        try:
            return decompose_newstyle_name(filename)[2:4]
        except Exception:
            # NOTE: load_mapping more conservative than fetch_mapping used in properties_from_mapping
            mapping = rmap.load_mapping(filename)
            return mapping.instrument, mapping.filekind
    elif config.is_reference(filename):
        result = get_reference_properties(filename)[2:4]
    else:
        try:
            result = properties_inside_mapping(filename)
        except Exception as exc:
            result = get_reference_properties(filename)[2:4]
    assert result[0] in INSTRUMENTS+[""], "Bad instrument " + \
        repr(result[0]) + " in filename " + repr(filename)
    assert result[1] in FILEKINDS+[""], "Bad filekind " + \
        repr(result[1]) + " in filename " + repr(filename)
    return result

def decompose_newstyle_name(filename):
    """
    >> decompose_newstyle_name("./hst.pmap")
    ('.', 'hst', '', '', '', '.pmap')

    >> decompose_newstyle_name("./hst_0001.pmap")
    ('.', 'hst', '', '', '0001', '.pmap')

    >> decompose_newstyle_name("./hst_acs.imap")
    ('.', 'hst', 'acs', '', '', '.imap')

    >> decompose_newstyle_name("./hst_acs_0001.imap")
    ('.', 'hst', 'acs', '', '0001', '.imap')

    >> decompose_newstyle_name("./hst_acs_biasfile.rmap")
    ('.', 'hst', 'acs', 'biasfile', '', '.rmap')

    >> decompose_newstyle_name("./hst_acs_biasfile_0001.rmap")
    ('.', 'hst', 'acs', 'biasfile', '0001', '.rmap')

    >> decompose_newstyle_name("./hst_acs_biasfile.fits")
    ('.', 'hst', 'acs', 'biasfile', '', '.fits')

    >> decompose_newstyle_name("./hst_acs_biasfile_0001.fits")
    ('.', 'hst', 'acs', 'biasfile', '0001', '.fits')
    """
    path, parts, ext = _get_fields(filename)
    observatory = parts[0]
    serial = list_get(parts, 3, "")

    if ext == ".pmap":
        assert len(parts) in [1,2], "Invalid .pmap filename " + repr(filename)
        instrument, filekind = "", ""
        serial = list_get(parts, 1, "")
    elif ext == ".imap":
        assert len(parts) in [2,3], "Invalid .imap filename " + repr(filename)
        instrument = parts[1]
        filekind = ""
        serial = list_get(parts, 2, "")
    else:
        assert len(parts) in [3,4], "Invalid filename " + repr(filename)
        instrument = parts[1]
        filekind = parts[2]
        serial = list_get(parts, 3, "")

    # Don't include filename in these or it messes up crds.certify unique error tracking.

    assert instrument in INSTRUMENTS+[""], "Invalid instrument " + repr(instrument)
    assert filekind in FILEKINDS+[""], "Invalid filekind " + repr(filekind)

    # The original regular expresssion "\d*\" matches anything in the below assert statement, so it's basically a no-op.
    # The correct way to do it is with fullmatch. However, that generates a bunch of test failures that need
    # to be reconciled before making the change:
    #     ERROR: test_throughput_lookup_generation (crds.tests.test_synphot_lookup_generator.TestSynphotLookupGenerator)
    #     FAIL: Doctest: crds.tests.test_bad_files.dt_bad_references_fast_mode
    #     FAIL: Doctest: crds.tests.test_bad_files.dt_bad_rules_jwst_getreferences_warning
    #     FAIL: Doctest: crds.tests.test_certify.certify_recursive
    #     FAIL: Doctest: crds.tests.test_certify.certify_table_comparison_context
    #     FAIL: Doctest: crds.tests.test_heavy_client.dt_getreferences_ignore_cache
    #     FAIL: Doctest: crds.tests.test_list.dt_list_cached_references
    #     FAIL: Doctest: crds.tests.test_synphot_hst.dt_synphot_core_integration_test
    #
    # XXXX TODO: Change to us fullmatch and resolve all the ensuing test failures.
    assert re.match(r"\d*", serial), "Invalid id field " + repr(id)
    #assert re.fullmatch(r"\d*", serial), "Invalid id field " + repr(id)
    # extension may vary for upload temporary files.

    return path, observatory, instrument, filekind, serial, ext

def properties_inside_mapping(filename):
    """Load `filename`s mapping header to discover and
    return (instrument, filekind).
    """
    map = rmap.fetch_mapping(filename)
    if map.mapping == "pipeline":
        result = "", ""
    elif map.mapping == "instrument":
        result = map.instrument, ""
    else:
        result = map.instrument, map.filekind
    return result

def _get_fields(filename):
    path = os.path.dirname(filename)
    name = os.path.basename(filename)
    name, ext = os.path.splitext(name)
    parts = name.split("_")
    return path, parts, ext

def list_get(l, index, default):
    try:
        return l[index]
    except IndexError:
        return default

def get_reference_properties(filename):
    """Figure out FITS (instrument, filekind, serial) based on `filename`.
    """
    try:   # Hopefully it's a nice new standard filename, easy
        return decompose_newstyle_name(filename)
    except AssertionError:  # cryptic legacy paths & names, i.e. reality
        pass
    # If not, dig inside the FITS file, slow
    return ref_properties_from_header(filename)

# =======================================================================

FILEKIND_KEYWORDS = ["REFTYPE", "TYPE", "META.REFTYPE",]

def ref_properties_from_header(filename):
    """Look inside FITS `filename` header to determine instrument, filekind.
    """
    # For legacy files,  just use the root filename as the unique id
    path, parts, ext = _get_fields(filename)
    serial = os.path.basename(os.path.splitext(filename)[0])
    header = data_file.get_free_header(filename, (), None, "jwst")
    header["TELESCOP"] = header["TELESCOPE"] = header["META.TELESCOPE"] = "jwst"
    name = os.path.basename(filename)
    try:
        instrument = utils.header_to_instrument(header).lower()
        assert instrument in INSTRUMENTS, "Invalid instrument " + repr(instrument)
    except Exception as exc:
        raise exceptions.CrdsNamingError(
            "Can't identify instrument of", repr(name), ":", str(exc)) from exc
    try:
        filekind = utils.get_any_of(header, FILEKIND_KEYWORDS, "UNDEFINED").lower()
        assert filekind in FILEKINDS, "Invalid file type " + repr(filekind)
    except Exception as exc:
        raise exceptions.CrdsNamingError("Can't identify REFTYPE of", repr(name))
    return path, "jwst", instrument, filekind, serial, ext

# =============================================================================

def reference_keys_to_dataset_keys(rmapping, header):
    """Given a header dictionary for a reference file, map the header back to keys
    relevant to datasets.  So for ACS biasfile the reference says BINAXIS1 but
    the dataset says NUMCOLS.  This would convert { "BINAXIS1": 1024 } to {
    "NUMCOLS" : 1024 }.

    In general,  rmap parkeys are matched against datset values and are defined
    as dataset header keywords.   For refactoring though,  what's initially
    available are reference file keywords...  which need to be mapped into the
    terms rmaps know:  dataset keywords.
    """
    header = dict(header)

    # Basic common pattern translations
    translations = {
            "META.EXPOSURE.P_EXPTYPE" : "META.EXPOSURE.TYPE",
            "P_EXP_TY" : "META.EXPOSURE.TYPE",

            "META.INSTRUMENT.P_BAND" : "META.INSTRUMENT.BAND",
            "P_BAND" : "META.INSTRUMENT.BAND",

            "META.INSTRUMENT.P_DETECTOR"  : "META.INSTRUMENT.DETECTOR",
            "P_DETECT"  : "META.INSTRUMENT.DETECTOR",

            "META.INSTRUMENT.P_CHANNEL" : "META.INSTRUMENT.CHANNEL",
            "P_CHANNE" : "META.INSTRUMENT.CHANNEL",

            "META.INSTRUMENT.P_FILTER" : "META.INSTRUMENT.FILTER",
            "P_FILTER"  : "META.INSTRUMENT.FILTER",

            "META.INSTRUMENT.P_PUPIL"  : "META.INSTRUMENT.PUPIL",
            "P_PUPIL" : "META.INSTRUMENT.PUPIL",

            "META.INSTRUMENT.P_MODULE"  : "META.INSTRUMENT.MODULE",
            "P_MODULE" : "META.INSTRUMENT.MODULE",

            "META.SUBARRAY.P_SUBARRAY" : "META.SUBARRAY.NAME",
            "P_SUBARR" : "META.SUBARRAY.NAME",

            "META.INSTRUMENT.P_GRATING" : "META.INSTRUMENT.GRATING",
            "P_GRATIN" : "META.INSTRUMENT.GRATING",

            "META.EXPOSURE.PREADPATT" : "META.EXPOSURE.READPATT",
            "META.EXPOSURE.P_READPATT" : "META.EXPOSURE.READPATT",
            "P_READPA" : "META.EXPOSURE.READPATT",

            # vvvv Speculative,  not currently defined or required by CAL vvvvv
            "META.INSTRUMENT.PCORONAGRAPH" : "META.INSTRUMENT.CORONAGRAPH",
            "P_CORONM" : "META.INSTRUMENT.CORONAGRAPH",
        }

    # Rmap header reference_to_dataset field tranlations,  can override basic!
    try:
        translations.update(rmapping.reference_to_dataset)
    except AttributeError:
        pass

    log.verbose("reference_to_dataset translations:\n", log.PP(translations), verbosity=60)
    log.verbose("reference_to_dataset input header:\n", log.PP(header), verbosity=80)

    for key in header:
        # Match META.X.P_SOMETHING or P_SOMETH
        if (key.split(".")[-1].startswith("P_")) and key not in translations:
            log.warning("CRDS-pattern-like keyword", repr(key),
                        "w/o CRDS translation to corresponding dataset keyword.")
            log.info("Pattern-like keyword", repr(key),
                     "may be misspelled or missing its translation in CRDS.  Pattern will not be used.")
            log.info("The translation for", repr(key),
                     "can be defined in crds.jwst.locate or rmap header reference_to_dataset field.")
            log.info("If this is not a pattern keyword, adding a translation to 'not-a-pattern'",
                     "will suppress this warning.")

    # Add replacements for translations *if* the existing untranslated value
    # is poor and the translated value is better defined.   This is to do
    # translations w/o replacing valid/concrete DM values with something
    # like guessed values of "UNDEFINED" or "N/A".
    for rkey in sorted(translations):
        if rkey in header:
            dkey = translations[rkey]
            dval = header.get(translations[rkey], None)
            rval = header[rkey]
            if rval not in [None, "UNDEFINED"] and rval != dval:
                log.info("Setting", repr(dkey), "=", repr(dval),
                         "to value of", repr(rkey), "=", repr(rval))
                header[dkey] = rval

    header = abstract.cross_strap_header(header)

    # NOTE:  the hacks below happen after cross-strapping and pattern handling
    # so if the keywords are still undefined they're undefined.  They have to
    # be explicitly defined as UNDEFINED somehow since they're nearly universally
    # used in constraints as condition variables even if they're not used in rmaps.
    # Unlike the targets of constraints,  CRDS is nominally unaware of condition
    # variables so they need to be incidentally defined.  This currently doesn't
    # work out if the rmap doesn't use them.  Condition variables are eval'ed in
    # expressions.

    if "SUBARRAY" not in header:
        header["SUBARRAY"] = header["META.SUBARRAY.NAME"] = "UNDEFINED"

    if "EXP_TYPE" not in header:
        header["EXP_TYPE"] = header["META.EXPOSURE.TYPE"] = "UNDEFINED"

    if "USEAFTER" not in header and "META.USEAFTER" in header:
        header["USEAFTER"] = header["META.USEAFTER"]
    if "USEAFTER" not in header and "META.USEAFTER" in header:
        header["USEAFTER"] = header["META.USEAFTER"]

    # If USEAFTER is defined,  or we're configured to fake it...
    #   don't invent one if its missing and we're not faking it.
    if "USEAFTER" in header or config.ALLOW_BAD_USEAFTER:

        # Identify this as best as possible,
        filename = header.get("FILENAME", None) or rmapping.filename

        reformatted = timestamp.reformat_useafter(filename, header).split()
        header["DATE-OBS"] = header["META.OBSERVATION.DATE"] = reformatted[0]
        header["TIME-OBS"] = header["META.OBSERVATION.TIME"] = reformatted[1]

    log.verbose("reference_to_dataset output header:\n", log.PP(header), verbosity=80)

    return header

# =============================================================================

def condition_matching_header(rmapping, header):
    """Normalize header values for .rmap reference insertion."""
    return dict(header)   # NOOP for JWST,  may have to revisit

# ============================================================================

def fits_to_parkeys(fits_header):
    """Map a FITS header onto rmap parkeys appropriate for JWST."""
    parkeys = {}
    for key, value in fits_header.items():
        key, value = str(key), str(value)
        if not key.lower().startswith("meta."):
            pk = cached_dm_find_fits_keyword(key)
            if not pk:
                pk = key
            else:
                assert len(pk) == 1, "CRDS JWST Data Model ambiguity on " + \
                    repr(key) + " = " + repr(pk)
                pk = pk[0]
        else:
            pk = key
        pk = pk.upper()
        parkeys[pk] = value
    return parkeys

@utils.cached
def cached_dm_find_fits_keyword(key):
    """Return the SSB JWST data model path for the specified non-path keyword,  nominally
    a FITS or json or ASDF bare keyword.
    """
    get_datamodels()
    return MODEL.find_fits_keyword(key.upper(), return_result=True)
# ============================================================================

def get_env_prefix(instrument):
    """Return the environment variable prefix (IRAF prefix) for `instrument`."""
    return "crds://"

# META.REF_FILE.SPECWCS.NAME.FITS_KEYWORD

def filekind_to_keyword(filekind):
    """Return the FITS keyword at which a reference should be recorded.

    >>> filekind_to_keyword("flat")
    'R_FLAT'
    >>> filekind_to_keyword("superbias")
    'R_SUPERB'
    """
    from . import schema
    flat_schema = schema.get_flat_schema()
    filekind = filekind.upper()
    meta_path = "META.REF_FILE.{}.NAME.FITS_KEYWORD".format(filekind)
    try:
        return flat_schema[meta_path]
    except KeyError:
        warn_filekind_once(filekind)
        return filekind

@utils.cached
def warn_filekind_once(filekind):
    log.warning("No apparent JWST cal code data models schema support for", log.srepr(filekind))

def locate_file(refname, mode=None):
    """Given a valid reffilename in CDBS or CRDS format,  return a cache path for the file.
    The aspect of this which is complicated is determining instrument and an instrument
    specific sub-directory for it based on the filename alone,  not the file contents.
    """
    if mode is  None:
        mode = config.get_crds_ref_subdir_mode(observatory="jwst")
    if mode == "instrument":
        instrument = utils.file_to_instrument(refname)
        rootdir = locate_dir(instrument, mode)
    elif mode == "flat":
        rootdir = config.get_crds_refpath("jwst")
    else:
        raise ValueError("Unhandled reference file location mode " + repr(mode))
    return  os.path.join(rootdir, os.path.basename(refname))

def locate_dir(instrument, mode=None):
    """Locate the instrument specific directory for a reference file."""
    if mode is  None:
        mode = config.get_crds_ref_subdir_mode(observatory="jwst")
    else:
        config.check_crds_ref_subdir_mode(mode)
    crds_refpath = config.get_crds_refpath("jwst")
    if mode == "instrument":   # use simple names inside CRDS cache.
        rootdir = os.path.join(crds_refpath, instrument.lower())
        if not os.path.exists(rootdir):
            if config.writable_cache_or_verbose("Skipping making instrument directory link for", repr(instrument)):
                utils.ensure_dir_exists(rootdir + "/locate_dir.fits")
    elif mode == "flat":    # use original flat cache structure,  all instruments in same directory.
        rootdir = crds_refpath
    else:
        raise ValueError("Unhandled reference file location mode " + repr(mode))
    return rootdir

# =======================================================================

# When loading headers,  make sure each keyword in a tuple is represented with
# the same value enabling any form to be used.  Case insensitive.

# XXXX the first translation should be the FITS keyword assuming there is one!!

CROSS_STRAPPED_KEYWORDS = {

    # META.REF_FILE.X is now obsolete but retained for backward compatibility.
    # it was replaced by META.X

    # These include non-DM keywords
    "META.INSTRUMENT.NAME" : ["INSTRUME", "INSTRUMENT", "META.INSTRUMENT.TYPE",],
    "META.TELESCOPE" : ["TELESCOP","TELESCOPE","META.TELESCOPE"],
    "META.DESCRIPTION" : ["DESCRIP","DESCRIPTION"],
    "META.REFTYPE" : ["REFTYPE",],

    # These include non-core-DM DM fields
    "META.AUTHOR" : ["AUTHOR",],
    "META.PEDIGREE" : ["PEDIGREE"],
    "META.USEAFTER" : ["USEAFTER"],
    "META.HISTORY" : ["HISTORY"],
    "META.CALIBRATION_SOFTWARE_VERSION" : ["CAL_VER", "CALIBRATION_SOFTWARE_VERSION"],
    "META.OBSERVATION.DATE" : ["DATE-OBS"],
    "META.OBSERVATION.TIME" : ["TIME-OBS"],


    # These should all be stock DM:FITS,  automatic
    # "META.INSTRUMENT.BAND" : ["BAND"],
    # "META.INSTRUMENT.CHANNEL" : ["CHANNEL"],
    # "META.INSTRUMENT.DETECTOR" : ["DETECTOR"],
    # "META.INSTRUMENT.FILTER" : ["FILTER"],
    # "META.INSTRUMENT.PUPIL" : ["PUPIL"],
    # "META.INSTRUMENT.GRATING" : ["GRATING"],

    # "META.SUBARRAY.NAME" : ["SUBARRAY"],
    # "META.SUBARRAY.XSTART" : ["SUBSTRT1"],
    # "META.SUBARRAY.YSTART" : ["SUBSTRT2"],
    # "META.SUBARRAY.XSIZE" : ["SUBSIZE1"],
    # "META.SUBARRAY.YSIZE" : ["SUBSIZE2"],
    # "META.SUBARRAY.FASTAXIS" : ["FASTAXIS"],
    # "META.SUBARRAY.SLOWAXIS" : ["SLOWAXIS"],

    # "META.EXPOSURE.TYPE" : ["EXP_TYPE"],
    # "META.EXPOSURE.READPATT" : ["READPATT"],

    # "META.APERTURE.NAME" : ["APERTURE", "APERNAME"],
}

# ============================================================================

@utils.cached
def get_static_pairs():
    return abstract.equivalence_dict_to_pairs(CROSS_STRAPPED_KEYWORDS)

def get_cross_strapped_pairs(header):
    """Return the list of keyword pairs where each pair describes synonyms for the same
    piece of data.
    """
    return  get_static_pairs() + _get_fits_datamodel_pairs(header)

def _get_fits_datamodel_pairs(header):
    """Return the (FITS, DM) and (DM, FITS) cross strap pairs associated with
    every keyword in `header` as defined by the datamodels interface functions
    defined by the CRDS JWST schema module.
    """
    pairs = []
    from . import schema
    for key in header:
        with log.verbose_warning_on_exception("Failed cross strapping keyword", repr(key)):
            fitskey = schema.dm_to_fits(key) or key
            dmkey = schema.fits_to_dm(key) or key
            pairs.append((fitskey, dmkey))
            pairs.append((dmkey, fitskey))
    log.verbose("Cal code datamodels keyword equivalencies:\n", log.PP(pairs), verbosity=90)
    return pairs


# ============================================================================

# Standard model names
DATA_MODEL_RE_STR = r"(META(\.[A-Z][A-Z0-9_]*)+)"
DATA_MODEL_RE = re.compile(DATA_MODEL_RE_STR)

def add_fits_keywords(log_message):
    """Process log `message` and annotate data model keywords/paths with
    their FITS translations if possible like:

    <data_models_keyword> '[' <fits_keyword> ']'
    """
    try:
        return _add_fits_keywords(log_message)
    except Exception:  # if annotation fails,  just return the original message
        return log_message

def _add_fits_keywords(log_message):
    """Process log `message` and annotate data model keywords/paths with
    their FITS translations if possible like:

    <data_models_keyword> '[' <fits_keyword> ']'
    """
    matches = DATA_MODEL_RE.finditer(log_message)
    if not matches:
        return log_message
    for dm_match in matches:
        model_key = dm_match.group(0)
        fits_key = schema.dm_to_fits(model_key)
        if fits_key is None:
            fits_key = _hack_fits_translation(model_key)
        annotation = " [" + fits_key + "]"
        if annotation not in log_message:
            log_message = log_message.replace(
                model_key, model_key + annotation)
    return log_message

# P_ keyword model names
DATA_MODEL_P_RE_STR = r"META(\.[A-Z][A-Z0-9_]*)*\.(P_[A-Z0-9_]+)"
DATA_MODEL_P_RE = re.compile(DATA_MODEL_P_RE_STR)

def _hack_fits_translation(model_key):
    """Hack FITS translations for data models keyword/paths not covered
    by dm_to_fits(), currently anything outside the core schema.
    """
    if model_key in CROSS_STRAPPED_KEYWORDS:
        fits_key = CROSS_STRAPPED_KEYWORDS[model_key][0]
    elif DATA_MODEL_P_RE.match(model_key):
        match = DATA_MODEL_P_RE.match(model_key)
        fits_key = match.group(2)[:8]
    else:
        fits_key = "FITS unknown"
    return fits_key

log.append_crds_filter(add_fits_keywords)

def disable_fits_annotations():
    log.remove_crds_filter(add_fits_keywords)

# ============================================================================

def test():
    """Run the module doctests."""
    import doctest
    from . import locate
    return doctest.testmod(locate)



"""CheckingProxy derived from jsonrpc.proxy due to subclassing problems
w/getattr. Converts service errors into ServiceError exceptions,  otherwise
call returns the jsonrpc "result" field.
"""
import sys
import uuid
import json
import time
import os

from urllib import request
import html
import gzip
import base64

# import crds
from crds.core import exceptions, log, config

# ============================================================================

def init_urlopen():
    """Call urlopen() once at import time to prepare for possible calls within
    multiprocessing processes.  This is magic which avoids a segfault on OS-X
    when urlopen() is called for the first time in a subprocess.

    %time showed this at around 2 msec on OS-X 10.11.67 El Capitain,  fine for
    a CRDS client unconditional one-off.
    """
    try:
        request.urlopen('')
    except Exception:
        pass

init_urlopen()

# ============================================================================

def apply_with_retries(func, *pars, **keys):
    """Apply function func() as f(*pargs, **keys) and return the result. Retry on any exception as defined in config.py"""
    retries = config.get_client_retry_count()
    delay = config.get_client_retry_delay_seconds()
    for retry in range(retries):
        try:
            return func(*pars, **keys)
        except Exception as exc:
            log.verbose_warning("FAILED: Attempt", str(retry+1), "of", retries, "with:", str(exc))
            log.verbose_warning("FAILED: Waiting for", delay, "seconds before retrying")  # waits after total fail...
            time.sleep(delay)
            exc2 = exc
    raise exc2

def message_id():
    """Return a nominal identifier for this program."""
    import crds
    return _program_name() + "-" + crds.__version__ + "-" + _PROCESS_ID + "-" + _request_id()

def _program_name():
    """Return the name of this program."""
    return os.path.basename(os.path.splitext(sys.argv[0])[0])

try:
    _PROCESS_ID = str(uuid.uuid4())
except Exception:
    _PROCESS_ID = "00000000-0000-0000-00000000000000000"

MSG_NO = 0
def _request_id():
    """Return an identifier unique to this particular JSONRPC request."""
    global MSG_NO
    MSG_NO += 1
    return "%08x" % MSG_NO

class CheckingProxy:
    """CheckingProxy converts calls to undefined methods into JSON RPC service
    calls bindings.   If the JSON rpc returns an error,  CheckingProxy raises a
    ServiceError exception containing the error's message.

    XXX NOTE: Always underscore new methods or you may hide a real JSONRPC method
    which also appears in the proxy object's namespace with the same name.

    """
    def __init__(self, service_url, version='1.0'):
        self.__version = str(version)
        self.__service_url = service_url

    def __getattr__(self, name):
        """Return a callable corresponding to JSONRPC method `name`."""
        return ServiceCallBinding(self.__service_url, name, self.__version)

    def __repr__(self):
        return self.__class__.__name__ + "(url='%s', version='%s')" % \
            (self.__service_url, self.__version)

class ServiceCallBinding:
    """When called,  ServiceCallBinding issues a JSONRPC call to the associated
    service URL.
    """
    def __init__(self, service_url, service_name=None, version='1.0'):
        self.__version = str(version)
        self.__service_url = service_url
        self.__service_name = service_name

    def __repr__(self):
        return self.__class__.__name__ + "(url='%s', method='%s')" % \
            (self.__service_url, self.__service_name)

    def _call(self, *args, **kwargs):
        """Core of RPC dispatch without error interpretation, logging, or return value decoding."""
        params = kwargs if len(kwargs) else args
        jsonrpc_params = {"jsonrpc": self.__version,
                          "method": self.__service_name,
                          'params': params,
                          'id': message_id()
                         }

        parameters = json.dumps(jsonrpc_params)

        url = self._get_url(jsonrpc_params)

        if "serverless" in url or "server-less" in url:
            raise exceptions.ServiceError("Configured for server-less mode.  Skipping JSON RPC " + repr(self.__service_name))

        if log.get_verbose() <= 50:
            log.verbose("CRDS JSON RPC", self.__service_name, params if len(str(params)) <= 60 else "(...)", "-->")
        else:
            log.verbose("CRDS JSON RPC to", url, "parameters", params, "-->")

        response = apply_with_retries(self._call_service, parameters, url)

        try:
            rval = json.loads(response)
        except Exception as exc:
            log.warning("Invalid CRDS jsonrpc response:\n", response)
            raise

        return rval

    def _get_url(self, jsonrpc_params):
        """Return the JSONRPC URL used to perform a method call.   Since post parameters are not visible in the
        log,  annotate the URL with additional method id paths which are functionally ignored but visible in
        the log.
        """
        return self.__service_url + jsonrpc_params["method"] + "/" + jsonrpc_params["id"] + "/"

    def _call_service(self, parameters, url):
        """Call the JSONRPC defined by `parameters` and raise a ServiceError on any exception."""
        if not isinstance(parameters, bytes):
            parameters = parameters.encode("utf-8")
        try:
            channel = request.urlopen(url, parameters)
            return channel.read().decode("utf-8")
        except Exception as exc:
            raise exceptions.ServiceError("CRDS jsonrpc failure " + repr(self.__service_name) + " " + str(exc)) from exc

    def __call__(self, *args, **kwargs):
        jsonrpc = self._call(*args, **kwargs)
        if jsonrpc["error"]:
            decoded = html.unescape(jsonrpc["error"]["message"])
            raise self.classify_exception(decoded)
        else:
            result = crds_decode(jsonrpc["result"])
            if isinstance(result, (str,int,float,bool)):
                log.verbose("RPC OK -->", repr(result))
            else:
                log.verbose("RPC OK", log.PP(result) if log.get_verbose() >= 75 else "")
            return result

    def classify_exception(self, decoded):
        """Interpret exc __str__ to define as more precise CRDS exception."""
        if "Channel" in decoded and "not found" in decoded:
            return exceptions.StatusChannelNotFoundError(decoded)
        elif "External agent requested calling process termination." in decoded:
            return exceptions.OwningProcessAbortedError(decoded)
        else:
            msg = "CRDS jsonrpc failure " + repr(self.__service_name) + " " + str(decoded)
            return exceptions.ServiceError(msg)

# ============================================================================

# These operate transparently in the proxy and are optionally used by the server.
#
# This makes a new client with crds_decoder compatible with both encoding and
# unencoding servers.
#
# An older client without crds_decoder will not work with a new server which is encoding.
# That could be achieved,  but wasn't because the function where the feature was
# needed would not work without compression anyway.

def crds_encode(obj):
    """Return a JSON-compatible encoding of `obj`,  nominally json-ified, compressed,
    and base64 encooded.   This is nominally to be called on the server.

    >>> obj = dict(p1="this", p2="that")
    >>> msg = crds_encode(obj)
    >>> isinstance(msg["crds_payload"], str)
    True
    """
    json_str = json.dumps(obj)
    utf8 = json_str.encode()
    compressed = gzip.compress(utf8)
    b64 = base64.b64encode(compressed)
    ascii = b64.decode("ascii")
    return dict(crds_encoded = "1.0",
                crds_payload = ascii)

def crds_decode(msg):
    """Decode something which was crds_encode'd,  or return it unaltered if
    it wasn't.
    
    >>> obj = dict(p1="this", p2="that")
    >>> msg = crds_encode(obj)
    >>> crds_decode(msg)
    {'p1': 'this', 'p2': 'that'}
    """
    if isinstance(msg, dict) and "crds_encoded" in msg:
        ascii = msg["crds_payload"]
        b64 = ascii.encode("ascii")
        compressed = base64.b64decode(b64)
        utf8 = gzip.decompress(compressed)
        json_str = utf8.decode()
        obj = json.loads(json_str)
        return obj
    else:
        return msg

def main():
    import doctest
    from crds.client import proxy
    return doctest.testmod(proxy)

if __name__ == "__main__":
    print(main())



"""This module contains doctests and unit tests which exercise some of the more
complex features of the basic rmap infrastructure.
"""
import os
import json
from pprint import pprint as pp
import pickle

from crds import rmap, log, config, tests, utils
from crds.client import api
from crds.exceptions import *
from crds.tests import test_config

from nose.tools import assert_raises, assert_true

# ==================================================================================

def dt_get_derived_from_created():
    """
    >>> old_state = test_config.setup()
    >>> p = rmap.get_cached_mapping("hst.pmap")
    >>> p.get_derived_from()
    CRDS - INFO - Skipping derivation checks for root mapping 'hst.pmap' derived_from = 'created by hand 12-23-2011'
    >>> test_config.cleanup(old_state)
    """
def dt_get_derived_from_phony():
    """
    >>> old_state = test_config.setup()
    >>> r = rmap.get_cached_mapping("data/hst_acs_darkfile_phony_derive.rmap")
    >>> r.get_derived_from()
    CRDS - WARNING - Parent mapping for 'hst_acs_darkfile_phony_derive.rmap' = 'phony.rmap' does not exist.
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_missing_references():
    """
    These are all missing because there is no reference file cache in this mode.

    >>> old_state = test_config.setup()
    >>> r = rmap.get_cached_mapping("data/hst_acs_darkfile_comment.rmap")
    >>> pp(r.missing_references())
    ['lcb12060j_drk.fits',
     'n3o1022cj_drk.fits',
     'n3o1022ej_drk.fits',
     'n3o1022fj_drk.fits',
     'n3o1022hj_drk.fits',
     'n3o1022ij_drk.fits',
     'n3o1022kj_drk.fits',
     'n3o1022lj_drk.fits',
     'r1u1415ij_drk.fits',
     'r1u1415kj_drk.fits',
     'r1u1415mj_drk.fits']
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_minimum_header():
    """
    >>> old_state = test_config.setup()
    >>> p = rmap.get_cached_mapping("hst.pmap")
    >>> pp(p.get_minimum_header("data/cos_N8XTZCAWQ.fits"))
    {'CAMERA': '3.0',
     'DATE-OBS': '2005-05-13',
     'FILTER': 'F160W',
     'INSTRUME': 'NICMOS',
     'NREAD': '1.0',
     'OBSMODE': 'MULTIACCUM',
     'READOUT': 'FAST',
     'REFTYPE': 'UNDEFINED',
     'SAMP_SEQ': 'SPARS64',
     'TIME-OBS': '11:00:47'}
    >>> test_config.cleanup(old_state)
    """



def dt_rmap_str():
    """
    >>> old_state = test_config.setup()
    >>> r = rmap.get_cached_mapping("data/hst_cos_bpixtab_0252.rmap")
    >>> print(str(r), end="")
    header = {
        'derived_from' : 'hst_cos_bpixtab_0251.rmap',
        'filekind' : 'BPIXTAB',
        'instrument' : 'COS',
        'mapping' : 'REFERENCE',
        'name' : 'hst_cos_bpixtab_0252.rmap',
        'observatory' : 'HST',
        'parkey' : (('DETECTOR',), ('DATE-OBS', 'TIME-OBS')),
        'reffile_format' : 'TABLE',
        'reffile_required' : 'NONE',
        'reffile_switch' : 'NONE',
        'rmap_relevance' : 'ALWAYS',
        'sha1sum' : 'd2024dade52a406af70fcdf27a81088004d67cae',
    }
    <BLANKLINE>
    selector = Match({
        ('FUV',) : UseAfter({
            '1996-10-01 00:00:00' : 's7g1700dl_bpix.fits',
            '2009-05-11 00:00:00' : 'z1r1943fl_bpix.fits',
        }),
        ('NUV',) : UseAfter({
            '1996-10-01 00:00:00' : 's7g1700pl_bpix.fits',
            '2009-05-11 00:00:00' : 'uas19356l_bpix.fits',
        }),
    })
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_obs_package():
    """
    >>> old_state = test_config.setup()
    >>> p = rmap.get_cached_mapping("data/hst_acs_darkfile.rmap")
    >>> p.obs_package.__name__
    'crds.hst'
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_format_with_comment():
    '''
    >>> old_state = test_config.setup()
    >>> r = rmap.get_cached_mapping("data/hst_acs_darkfile_comment.rmap")
    >>> print(r.comment, end="")
    <BLANKLINE>
    This is a block comment which can be used to store additional metadata
    about the state and evolution of this type and files.

    >>> print(r, end="")
    header = {
        'derived_from' : 'generated from CDBS database 2013-01-11 13:58:14.664182',
        'filekind' : 'DARKFILE',
        'instrument' : 'ACS',
        'mapping' : 'REFERENCE',
        'name' : 'hst_acs_darkfile_comment.rmap',
        'observatory' : 'HST',
        'parkey' : (('DETECTOR', 'CCDAMP', 'CCDGAIN'), ('DATE-OBS', 'TIME-OBS')),
        'parkey_relevance' : {
            'ccdamp' : '(DETECTOR != "SBC")',
            'ccdgain' : '(DETECTOR != "SBC")',
        },
        'rmap_relevance' : 'ALWAYS',
        'sha1sum' : '0b3af86642812a1af65b77d429886e186acef915',
    }
    <BLANKLINE>
    comment = """
    This is a block comment which can be used to store additional metadata
    about the state and evolution of this type and files.
    """
    <BLANKLINE>
    selector = Match({
        ('HRC', 'A|ABCD|AD|B|BC|C|D', '1.0|2.0|4.0|8.0') : UseAfter({
            '1992-01-01 00:00:00' : 'lcb12060j_drk.fits',
            '2002-03-01 00:00:00' : 'n3o1022cj_drk.fits',
            '2002-03-18 00:00:00' : 'n3o1022ej_drk.fits',
            '2002-03-19 00:34:31' : 'n3o1022fj_drk.fits',
            '2002-03-20 00:34:32' : 'n3o1022hj_drk.fits',
            '2002-03-21 00:34:31' : 'n3o1022ij_drk.fits',
            '2002-03-22 00:34:30' : 'n3o1022kj_drk.fits',
            '2002-03-23 00:34:28' : 'n3o1022lj_drk.fits',
            '2007-01-21 02:09:05' : 'r1u1415ij_drk.fits',
            '2007-01-22 00:40:13' : 'r1u1415kj_drk.fits',
            '2007-01-26 00:07:33' : 'r1u1415mj_drk.fits',
        }),
    })
    >>> test_config.cleanup(old_state)
    '''

def dt_rmap_warn_checksum():
    """
    >>> old_state = test_config.setup()
    >>> r = rmap.ReferenceMapping.from_string('''
    ... header = {
    ...    'derived_from' : 'generated from CDBS database 2013-01-11 13:58:14.664182',
    ...    'filekind' : 'DARKFILE',
    ...    'instrument' : 'ACS',
    ...    'mapping' : 'REFERENCE',
    ...    'name' : 'hst_acs_darkfile_comment.rmap',
    ...    'observatory' : 'HST',
    ...    'parkey' : (('DETECTOR', 'CCDAMP', 'CCDGAIN'), ('DATE-OBS', 'TIME-OBS')),
    ...    'sha1sum' : "something bad",
    ... }
    ...
    ... selector = Match({
    ...    ('HRC', 'A|ABCD|AD|B|BC|C|D', '1.0|2.0|4.0|8.0') : UseAfter({
    ...        '1992-01-01 00:00:00' : 'lcb12060j_drk.fits',
    ...        '2002-03-01 00:00:00' : 'n3o1022cj_drk.fits',
    ...     }),
    ... })
    ... ''', ignore_checksum='warn')
    CRDS - WARNING - Checksum error : sha1sum mismatch in '(noname)'
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_get_reference_parkeys():
    """
    >>> old_state = test_config.setup()
    >>> r = rmap.get_cached_mapping("data/jwst_miri_specwcs_0004.rmap")
    >>> r.parkey
    (('META.INSTRUMENT.DETECTOR', 'META.INSTRUMENT.CHANNEL', 'META.INSTRUMENT.BAND', 'META.SUBARRAY.NAME'),)
    >>> r.get_reference_parkeys()
    ('BAND', 'CHANNEL', 'DETECTOR', 'META.EXPOSURE.TYPE', 'META.INSTRUMENT.BAND', 'META.INSTRUMENT.CHANNEL', 'META.INSTRUMENT.DETECTOR', 'META.INSTRUMENT.LAMP_STATE', 'META.SUBARRAY.NAME', 'META.VISIT.TSOVISIT', 'SUBARRAY')
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_get_valid_values_map():
    """
    >>> old_state = test_config.setup()
    >>> i = rmap.get_cached_mapping("hst_acs.imap")
    >>> pp(i.get_valid_values_map())
    {'APERTURE': ['NONE',
                  'SBC',
                  'SBC-FIX',
                  'WFC',
                  'WFC-FIX',
                  'WFC1',
                  'WFC1-1K',
                  'WFC1-2K',
                  'WFC1-512',
                  'WFC1-CTE',
                  'WFC1-FIX',
                  'WFC1-IRAMP',
                  'WFC1-IRAMPQ',
                  'WFC1-MRAMP',
                  'WFC1-MRAMPQ',
                  'WFC1-POL0UV',
                  'WFC1-POL0V',
                  'WFC1-POL120UV',
                  'WFC1-POL120V',
                  'WFC1-POL60UV',
                  'WFC1-POL60V',
                  'WFC1-SMFL',
                  'WFC1A-1K',
                  'WFC1A-2K',
                  'WFC1A-512',
                  'WFC1B-1K',
                  'WFC1B-2K',
                  'WFC1B-512',
                  'WFC2',
                  'WFC2-1K',
                  'WFC2-2K',
                  'WFC2-512',
                  'WFC2-FIX',
                  'WFC2-MRAMP',
                  'WFC2-MRAMPQ',
                  'WFC2-ORAMP',
                  'WFC2-ORAMPQ',
                  'WFC2-POL0UV',
                  'WFC2-POL0V',
                  'WFC2-POL120UV',
                  'WFC2-POL120V',
                  'WFC2-POL60UV',
                  'WFC2-POL60V',
                  'WFC2-SMFL',
                  'WFC2C-1K',
                  'WFC2C-2K',
                  'WFC2C-512',
                  'WFC2D-1K',
                  'WFC2D-2K',
                  'WFC2D-512',
                  'WFCENTER'],
     'ATODCORR': [],
     'BIASCORR': [],
     'CCDAMP': ['A', 'ABCD', 'AC', 'AD', 'B', 'BC', 'BD', 'C', 'D'],
     'CCDCHIP': [],
     'CCDGAIN': ['0.5', '1.0', '1.4', '2.0', '4.0', '8.0'],
     'CRCORR': [],
     'DARKCORR': [],
     'DETECTOR': ['HRC', 'SBC', 'WFC'],
     'DQICORR': [],
     'DRIZCORR': [],
     'FILTER1': ['BLOCK1',
                 'BLOCK2',
                 'BLOCK3',
                 'BLOCK4',
                 'CLEAR1L',
                 'CLEAR1S',
                 'F115LP',
                 'F122M',
                 'F125LP',
                 'F140LP',
                 'F150LP',
                 'F165LP',
                 'F475W',
                 'F502N',
                 'F550M',
                 'F555W',
                 'F606W',
                 'F625W',
                 'F658N',
                 'F775W',
                 'F850LP',
                 'F892N',
                 'G800L',
                 'POL0UV',
                 'POL120UV',
                 'POL60UV',
                 'PR110L',
                 'PR130L'],
     'FILTER2': ['CLEAR2L',
                 'CLEAR2S',
                 'F220M',
                 'F220W',
                 'F250W',
                 'F330W',
                 'F344N',
                 'F410W',
                 'F435W',
                 'F660N',
                 'F814W',
                 'FR1016N',
                 'FR388N',
                 'FR423N',
                 'FR459M',
                 'FR462N',
                 'FR505N',
                 'FR551N',
                 'FR555N',
                 'FR601N',
                 'FR647M',
                 'FR656N',
                 'FR716N',
                 'FR782N',
                 'FR853N',
                 'FR914M',
                 'FR931N',
                 'G800L',
                 'POL0V',
                 'POL120V',
                 'POL60V',
                 'PR200L'],
     'FLASHCUR': ['HIGH', 'LOW', 'MED'],
     'FLATCORR': [],
     'FLSHCORR': [],
     'FW1OFFST': ['-1', '0', '1', '2'],
     'FW2OFFST': ['-1', '0', '1', '2'],
     'FWSOFFST': ['-1', '0', '1', '2'],
     'GLINCORR': [],
     'LTV1': [],
     'LTV2': [],
     'NUMCOLS': [],
     'NUMROWS': [],
     'OBSTYPE': ['CORONAGRAPHIC', 'IMAGING', 'INTERNAL', 'SPECTROSCOPIC'],
     'PCTECORR': [],
     'PHOTCORR': [],
     'SHADCORR': [],
     'SHUTRPOS': ['A', 'B'],
     'XCORNER': [],
     'YCORNER': []}

    >>> pp(i.get_valid_values_map(remove_special=False)["CCDGAIN"])
    ['0.5', '1.0', '1.4', '2.0', '4.0', '8.0']

    >>> pp(i.get_valid_values_map(condition=True)["FW1OFFST"])
    ['-1.0', '0.0', '1.0', '2.0']
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_get_valid_values_map_range():
    """
    >>> old_state = test_config.setup()
    >>> r = rmap.get_cached_mapping("data/hst_wfpc2_flatfile.rmap")
    >>> r.get_valid_values_map() == {'FILTER1': ('0.0', '1.0', '2.0', '3.0', '4.0', '5.0', '6.0', '7.0', '8.0', '9.0', '10.0', '11.0', '12.0', '13.0', '14.0', '15.0', '16.0', '17.0', '18.0', '19.0', '20.0', '21.0', '22.0', '23.0', '24.0', '25.0', '26.0', '27.0', '28.0', '29.0', '30.0', '31.0', '32.0', '33.0', '34.0', '35.0', '36.0', '37.0', '38.0', '39.0', '40.0', '41.0', '42.0', '43.0', '44.0', '45.0', '46.0', '47.0', '48.0', '49.0', '50.0', '51.0', '52.0', '53.0', '54.0', '55.0', '56.0', '57.0', '58.0', '59.0', '60.0', '61.0', '62.0', '63.0', '64.0', '65.0', '66.0', '67.0', '68.0', '69.0', '70.0', '71.0'), 'FILTER2': ('0.0', '1.0', '2.0', '3.0', '4.0', '5.0', '6.0', '7.0', '8.0', '9.0', '10.0', '11.0', '12.0', '13.0', '14.0', '15.0', '16.0', '17.0', '18.0', '19.0', '20.0', '21.0', '22.0', '23.0', '24.0', '25.0', '26.0', '27.0', '28.0', '29.0', '30.0', '31.0', '32.0', '33.0', '34.0', '35.0', '36.0', '37.0', '38.0', '39.0', '40.0', '41.0', '42.0', '43.0', '44.0', '45.0', '46.0', '47.0', '48.0', '49.0', '50.0', '51.0', '52.0', '53.0', '54.0', '55.0', '56.0', '57.0', '58.0', '59.0', '60.0', '61.0', '62.0', '63.0', '64.0', '65.0', '66.0', '67.0', '68.0', '69.0', '70.0', '71.0'), 'MODE': ('FULL', 'AREA')}
    True
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_get_best_references_fail():
    """
    >>> old_state = test_config.setup()
    >>> i = rmap.get_cached_mapping("hst_acs.imap")
    >>> i.get_best_references({
    ... "DETECTOR" : "HRC",
    ... "CCDAMP" : "B",
    ... "CCDGAIN" : "7.0",
    ... "DARKCORR" : "PERFORM",
    ... "DATE-OBS" : "2015-04-30",
    ... "TIME-OBS" : "16:43:00",
    ... }, include=["darkfile"])
    {'darkfile': 'NOT FOUND No match found.'}
    >>> test_config.cleanup(old_state)
    """

def dt_validate_mapping_valid():
    """
    >>> old_state = test_config.setup()
    >>> r = rmap.get_cached_mapping("data/hst_acs_darkfile.rmap")
    >>> r.validate_mapping()
    >>> test_config.cleanup(old_state)
    """

def dt_validate_mapping_invalid1():
    """
    >>> old_state = test_config.setup()
    >>> r = rmap.get_cached_mapping("data/hst_acs_darkfile_invalid1.rmap")
    >>> r.validate_mapping()
    CRDS - ERROR - Match('DETECTOR', 'CCDAMP', 'CCDGAIN') : ('HRC', 'A|ABCD|AD|B|BC|C|DDDD', '1.0|2.0|4.0|8.0') :  parameter='CCDAMP' value='DDDD' is not in ('A', 'B', 'C', 'D', 'AC', 'AD', 'BC', 'BD', 'ABCD', 'N/A')
    >>> test_config.cleanup(old_state)
    """

def dt_validate_mapping_invalid2():
    """
    >>> old_state = test_config.setup()
    >>> r = rmap.get_cached_mapping("data/hst_acs_darkfile_invalid2.rmap")
    >>> r.validate_mapping()
    CRDS - ERROR - Match('DETECTOR', 'CCDAMP', 'CCDGAIN') : ('FOOBAR', 'A|ABCD|AD|B|BC|C|DDDD', '1.0|2.0|4.0|8.0') :  parameter='DETECTOR' value='FOOBAR' is not in ('WFC', 'HRC', 'SBC')
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_asmapping_readonly():
    """
    >>> old_state = test_config.setup()
    >>> r = rmap.asmapping("data/hst_acs_darkfile.rmap", cached="readonly")
    >>> test_config.cleanup(old_state)
    """

def dt_rmap_get_equivalent_mapping_missing():
    """
    >>> old_state = test_config.setup()
    >>> p = rmap.get_cached_mapping("hst.pmap")
    >>> p.get_equivalent_mapping("hst_cos_twozxtab_0001.rmap")
    CRDS - WARNING - No equivalent filekind in 'hst_cos.imap' corresponding to 'hst_cos_twozxtab_0001.rmap'
    >>> test_config.cleanup(old_state)
    """

def dt_imap_match_not_applicable():
    """
    >>> old_state = test_config.setup()
    >>> p = rmap.get_cached_mapping("data/hst_acs_9999.imap")
    >>> p.get_best_references({
    ...      "DETECTOR" : "SBC",
    ...      "CCDAMP" : "A",
    ...      "CCDGAIN" : "1.0",
    ...      "DATE-OBS" : "1993-01-01",
    ...      "TIME-OBS" : "12:00:00",
    ...      "OBSTYPE" : "IMAGING",
    ...      "FLATCORR" : "PERFORM",
    ...      "DQICORR" : "PERFORM",
    ...      "DRIZCORR" : "PERFORM",
    ...      "PHOTCORR" : "PERFORM",
    ...      "DRIZCORR" : "PERFORM",
    ... })["pctetab"]
    'NOT FOUND n/a'
    >>> test_config.cleanup(old_state)
    """

def dt_imap_match_omit():
    """
    >>> old_state = test_config.setup()
    >>> p = rmap.get_cached_mapping("data/hst_acs_9999.imap")
    >>> "mlintab" in p.get_best_references({
    ...      "DETECTOR" : "SBC",
    ...      "CCDAMP" : "A",
    ...      "CCDGAIN" : "1.0",
    ...      "DATE-OBS" : "2002-03-19",
    ...      "TIME-OBS" : "00:34:32",
    ...      "OBSTYPE" : "IMAGING",
    ...      "FLATCORR" : "PERFORM",
    ...      "DQICORR" : "PERFORM",
    ...      "DRIZCORR" : "PERFORM",
    ...      "PHOTCORR" : "PERFORM",
    ...      "DRIZCORR" : "PERFORM",
    ... })
    False
    >>> test_config.cleanup(old_state)
    """

def dt_pickling():
    """
    >>> from crds import data_file

    >>> old_state = test_config.setup()

    >>> p = rmap.get_cached_mapping("hst.pmap")

    >>> s = pickle.dumps(p)
    >>> q = pickle.loads(s)

    >>> str(q) == str(p)
    True

    >>> p.difference(q)
    []
    >>> q.difference(p)
    []

    >>> p.validate_mapping()
    >>> q.validate_mapping()

    >>> header = data_file.get_header("data/j8bt06o6q_raw.fits")
    >>> refs_p = p.get_best_references(header)
    >>> refs_q = q.get_best_references(header)
    >>> refs_p == refs_q
    True

    >>> test_config.cleanup(old_state)
    """

# ==================================================================================

class TestRmap(test_config.CRDSTestCase):

    def test_rmap_get_imap_except(self):
        r = rmap.get_cached_mapping("hst.pmap")
        with self.assertRaises(CrdsUnknownInstrumentError):
            r.get_imap("foo")

    def test_rmap_get_filekind(self):
        r = rmap.get_cached_mapping("hst.pmap")
        self.assertEqual(set(r.get_filekinds("data/j8bt05njq_raw.fits")),
                         {'atodtab','biasfile','bpixtab','ccdtab','cfltfile','crrejtab',
                          'd2imfile','darkfile','dgeofile','drkcfile','flshfile','idctab',
                          'imphttab','mdriztab','mlintab','npolfile','oscntab','pctetab',
                          'pfltfile','shadfile','spottab'})

    def test_rmap_get_equivalent_mapping(self):
        i = rmap.get_cached_mapping("data/hst_acs_0002.imap")
        self.assertEqual(i.get_equivalent_mapping("hst.pmap"), None)
        self.assertEqual(i.get_equivalent_mapping("data/hst_acs_0001.imap").name, "hst_acs.imap")
        self.assertEqual(i.get_equivalent_mapping("data/hst_acs_biasfile_0002.rmap").name, "hst_acs_biasfile.rmap")

    def test_rmap_list_mappings(self):
        os.environ["CRDS_MAPPATH_SINGLE"] = self.data_dir
        self.assertEqual(rmap.list_mappings("*.imap", "hst"), [
                'hst_acs.imap', 'hst_acs_0001.imap', 'hst_acs_0002.imap', 'hst_acs_9999.imap',
                'hst_cos.imap', 'hst_nicmos.imap', 'hst_stis.imap', 'hst_wfc3.imap', 'hst_wfpc2.imap',
                'jwst_fgs_na.imap', 'jwst_miri_omit.imap', 'jwst_niriss_na_omit.imap', 'roman_wfi_0001.imap'])

    def test_rmap_list_references(self):
        os.environ["CRDS_REFPATH_SINGLE"] = self.data_dir
        config.CRDS_REF_SUBDIR_MODE = "flat"
        self.assertEqual(rmap.list_references("*.r1h", "hst"), ['dbu1405fu.r1h', 'dbu1405iu.r1h', 'e1b09593u.r1h', 'e1b09594u.r1h', 'valid.r1h'])

    def test_rmap_get_derived_from(self):
        # api.dump_mappings("hst.pmap", mappings=["hst_acs_flshfile_0251.rmap"])
        os.environ["CRDS_MAPPATH_SINGLE"] = self.data_dir
        r = rmap.get_cached_mapping("data/hst_acs_flshfile_0252.rmap")
        self.assertEqual(r.get_derived_from().name, 'hst_acs_flshfile_0251.rmap')

    def test_missing_required_header_key(self):
        with self.assertRaises(MissingHeaderKeyError):
            r = rmap.load_mapping("data/hst_acs_darkfile_missing_key.rmap")

    def test_load_rmap_bad_expr(self):
        with self.assertRaises(SyntaxError):
            r = rmap.get_cached_mapping("data/hst_acs_darkfile_badexpr.rmap")

    def test_rmap_get_parkey_map(self):
        i = rmap.get_cached_mapping("hst_acs.imap")
        with self.assertRaises(CrdsUnknownReftypeError):
            i.get_rmap("foo")

    def test_rmap_missing_checksum(self):
        with self.assertRaises(ChecksumError):
            r = rmap.ReferenceMapping.from_string('''
header = {
   'derived_from' : 'generated from CDBS database 2013-01-11 13:58:14.664182',
   'filekind' : 'DARKFILE',
   'instrument' : 'ACS',
   'mapping' : 'REFERENCE',
   'name' : 'hst_acs_darkfile_comment.rmap',
   'observatory' : 'HST',
   'parkey' : (('DETECTOR', 'CCDAMP', 'CCDGAIN'), ('DATE-OBS', 'TIME-OBS')),
}

selector = Match({
   ('HRC', 'A|ABCD|AD|B|BC|C|D', '1.0|2.0|4.0|8.0') : UseAfter({
       '1992-01-01 00:00:00' : 'lcb12060j_drk.fits',
       '2002-03-01 00:00:00' : 'n3o1022cj_drk.fits',
    }),
})
''')

    def test_rmap_schema_uri(self):
        r = rmap.ReferenceMapping.from_string('''
header = {
    'derived_from' : 'jwst_nircam_pars-tweakregstep_0001.rmap',
    'file_ext' : '.asdf',
    'filekind' : 'pars-tweakregstep',
    'filetype' : 'pars-tweakregstep',
    'instrument' : 'NIRCAM',
    'mapping' : 'REFERENCE',
    'name' : 'jwst_nircam_pars-tweakregstep_0002.rmap',
    'observatory' : 'JWST',
    'parkey' : (('META.EXPOSURE.TYPE', 'META.INSTRUMENT.FILTER', 'META.INSTRUMENT.PUPIL'), ('META.OBSERVATION.DATE', 'META.OBSERVATION.TIME')),
    'schema_uri' : 'http://stsci.edu/schemas/asdf/core/ndarray-1.0.0',
    'sha1sum' : '186bb16c5b4ec498d9cc7d03ff564ae22f221d6f',
    'suffix' : 'pars-tweakregstep',
    'text_descr' : 'TweakRegStep runtime parameters',
}

selector = Match({
})
''', ignore_checksum=True)
        self.assertEquals(r.schema_uri, 'http://stsci.edu/schemas/asdf/core/ndarray-1.0.0')
        r.validate()

    def test_rmap_schema_uri_missing(self):
        r = rmap.ReferenceMapping.from_string('''
header = {
    'derived_from' : 'jwst_nircam_pars-tweakregstep_0001.rmap',
    'file_ext' : '.asdf',
    'filekind' : 'pars-tweakregstep',
    'filetype' : 'pars-tweakregstep',
    'instrument' : 'NIRCAM',
    'mapping' : 'REFERENCE',
    'name' : 'jwst_nircam_pars-tweakregstep_0002.rmap',
    'observatory' : 'JWST',
    'parkey' : (('META.EXPOSURE.TYPE', 'META.INSTRUMENT.FILTER', 'META.INSTRUMENT.PUPIL'), ('META.OBSERVATION.DATE', 'META.OBSERVATION.TIME')),
    'schema_uri' : 'http://stsci.edu/schemas/asdf/core/does_not_exist-1.0.0',
    'sha1sum' : '186bb16c5b4ec498d9cc7d03ff564ae22f221d6f',
    'suffix' : 'pars-tweakregstep',
    'text_descr' : 'TweakRegStep runtime parameters',
}

selector = Match({
})
''', ignore_checksum=True)
        with self.assertRaises(FileNotFoundError):
            r.validate()

    def test_rmap_get_best_references_include(self):
        r = rmap.get_cached_mapping("data/hst_acs_darkfile_comment.rmap")
        header = {
            'CCDAMP': 'ABCD',
            'CCDGAIN': '1.0',
            'DARKCORR': 'UNDEFINED',
            'DATE-OBS': '2002-07-18',
            'DETECTOR': 'WFC',
            'TIME-OBS': '18:09:15.773332'}
        with self.assertRaises(CrdsUnknownReftypeError):
            r.get_best_references(header, include=["flatfile"])

    def test_rmap_get_parkey_map(self):
        i = rmap.get_cached_mapping("hst_acs.imap")
        i.get_parkey_map() == {'APERTURE': ['*',
                  'WFC',
                  'WFC-FIX',
                  'WFC1',
                  'WFC1-1K',
                  'WFC1-2K',
                  'WFC1-512',
                  'WFC1-CTE',
                  'WFC1-FIX',
                  'WFC1-IRAMP',
                  'WFC1-IRAMPQ',
                  'WFC1-MRAMP',
                  'WFC1-MRAMPQ',
                  'WFC1-POL0UV',
                  'WFC1-POL0V',
                  'WFC1-POL120UV',
                  'WFC1-POL120V',
                  'WFC1-POL60UV',
                  'WFC1-POL60V',
                  'WFC1-SMFL',
                  'WFC2',
                  'WFC2-2K',
                  'WFC2-FIX',
                  'WFC2-MRAMP',
                  'WFC2-ORAMP',
                  'WFC2-ORAMPQ',
                  'WFC2-POL0UV',
                  'WFC2-POL0V',
                  'WFC2-POL120UV',
                  'WFC2-POL120V',
                  'WFC2-POL60UV',
                  'WFC2-POL60V',
                  'WFC2-SMFL',
                  'WFCENTER'],
     'ATODCORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'BIASCORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'CCDAMP': ['A', 'ABCD', 'AC', 'AD', 'B', 'BC', 'BD', 'C', 'D', 'N/A'],
     'CCDCHIP': ['N/A'],
     'CCDGAIN': ['0.5', '1.0', '1.4', '2.0', '4.0', '8.0', 'N/A'],
     'CRCORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'DARKCORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'DETECTOR': ['HRC', 'SBC', 'WFC'],
     'DQICORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'DRIZCORR': ['NONE', 'COMPLETE', 'PERFORM', 'OMIT', 'UNDEFINED'],
     'FILTER1': ['*',
                 'BLOCK1',
                 'BLOCK2',
                 'BLOCK3',
                 'BLOCK4',
                 'CLEAR1L',
                 'CLEAR1S',
                 'F115LP',
                 'F122M',
                 'F125LP',
                 'F140LP',
                 'F150LP',
                 'F165LP',
                 'F475W',
                 'F502N',
                 'F550M',
                 'F555W',
                 'F606W',
                 'F625W',
                 'F658N',
                 'F775W',
                 'F850LP',
                 'F892N',
                 'G800L',
                 'N/A',
                 'POL0UV',
                 'POL120UV',
                 'POL60UV',
                 'PR110L',
                 'PR130L'],
     'FILTER2': ['CLEAR2L',
                 'CLEAR2S',
                 'F220M',
                 'F220W',
                 'F250W',
                 'F330W',
                 'F344N',
                 'F410W',
                 'F435W',
                 'F660N',
                 'F814W',
                 'FR1016N',
                 'FR388N',
                 'FR423N',
                 'FR459M',
                 'FR462N',
                 'FR505N',
                 'FR551N',
                 'FR555N',
                 'FR601N',
                 'FR647M',
                 'FR656N',
                 'FR716N',
                 'FR782N',
                 'FR853N',
                 'FR914M',
                 'FR931N',
                 'N/A',
                 'POL0V',
                 'POL120V',
                 'POL60V',
                 'PR200L'],
     'FLASHCUR': ['HIGH', 'LOW', 'MED'],
     'FLATCORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'FLSHCORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'FW1OFFST': ['-1.0', '1.0', 'N/A'],
     'FW2OFFST': ['-1.0', '1.0', 'N/A'],
     'FWSOFFST': ['N/A'],
     'GLINCORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'LTV1': ['-1816.0',
              '-2048.0',
              '-3072.0',
              '-3584.0',
              '-3604.0',
              '19.0',
              '22.0',
              '24.0',
              'N/A'],
     'LTV2': ['-1.0',
              '-1023.0',
              '-1535.0',
              '-1591.0',
              '-57.0',
              '-824.0',
              '0.0',
              '20.0',
              'N/A'],
     'NUMCOLS': ['1046.0',
                 '1062.0',
                 '2070.0',
                 '2300.0',
                 '4144.0',
                 '512.0',
                 '534.0',
                 'N/A'],
     'NUMROWS': ['1024.0', '1044.0', '2046.0', '2068.0', '400.0', '512.0', 'N/A'],
     'OBSTYPE': ['CORONAGRAPHIC', 'IMAGING', 'SPECTROSCOPIC'],
     'PCTECORR': ['NONE', 'COMPLETE', 'PERFORM', 'OMIT', 'UNDEFINED'],
     'PHOTCORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'SHADCORR': ['PERFORM', 'NONE', 'OMIT', 'COMPLETE', 'UNDEFINED'],
     'SHUTRPOS': ['A', 'B'],
     'XCORNER': ['N/A'],
     'YCORNER': ['N/A']}

    maxDiff = None

    def test_rmap_todict(self):
        r = rmap.get_cached_mapping("data/hst_cos_bpixtab_0252.rmap")
        self.assertEqual(r.todict(), {'text_descr': 'Data Quality (Bad Pixel) Initialization Table', 'selections': [('FUV', '1996-10-01 00:00:00', 's7g1700dl_bpix.fits'), ('FUV', '2009-05-11 00:00:00', 'z1r1943fl_bpix.fits'), ('NUV', '1996-10-01 00:00:00', 's7g1700pl_bpix.fits'), ('NUV', '2009-05-11 00:00:00', 'uas19356l_bpix.fits')], 'header': {'sha1sum': 'd2024dade52a406af70fcdf27a81088004d67cae', 'reffile_switch': 'none', 'filekind': 'bpixtab', 'instrument': 'cos', 'derived_from': 'hst_cos_bpixtab_0251.rmap', 'reffile_format': 'table', 'observatory': 'hst', 'parkey': (('DETECTOR',), ('DATE-OBS', 'TIME-OBS')), 'reffile_required': 'none', 'rmap_relevance': 'always', 'mapping': 'reference', 'name': 'hst_cos_bpixtab_0252.rmap'}, 'parameters': ('DETECTOR', 'USEAFTER', 'REFERENCE')})

    def test_rmap_tojson(self):
        r = rmap.get_cached_mapping("data/hst_cos_bpixtab_0252.rmap")
        self.assertEqual(json.loads(r.tojson()), {u'header': {u'observatory': u'hst', u'name': u'hst_cos_bpixtab_0252.rmap', u'reffile_required': u'none', u'parkey': [[u'DETECTOR'], [u'DATE-OBS', u'TIME-OBS']], u'mapping': u'reference', u'filekind': u'bpixtab', u'instrument': u'cos', u'derived_from': u'hst_cos_bpixtab_0251.rmap', u'reffile_switch': u'none', u'reffile_format': u'table', u'rmap_relevance': u'always', u'sha1sum': u'd2024dade52a406af70fcdf27a81088004d67cae'}, u'text_descr': u'Data Quality (Bad Pixel) Initialization Table', u'parameters': [u'DETECTOR', u'USEAFTER', u'REFERENCE'], u'selections': [[u'FUV', u'1996-10-01 00:00:00', u's7g1700dl_bpix.fits'], [u'FUV', u'2009-05-11 00:00:00', u'z1r1943fl_bpix.fits'], [u'NUV', u'1996-10-01 00:00:00', u's7g1700pl_bpix.fits'], [u'NUV', u'2009-05-11 00:00:00', u'uas19356l_bpix.fits']]})

    def test_rmap_match_not_applicable(self):
        r = rmap.get_cached_mapping("data/hst_acs_darkfile_na_omit.rmap")
        r.get_best_ref({
                "DETECTOR" : "SBC",
                "CCDAMP" : "A",
                "CCDGAIN" : "1.0",
                "DATE-OBS" : "1993-01-01",
                "TIME-OBS" : "12:00:00",
                }) == "NOT FOUNT n/a"

    def test_rmap_match_omit(self):
        r = rmap.get_cached_mapping("data/hst_acs_darkfile_na_omit.rmap")
        r.get_best_ref({
                "DETECTOR" : "SBC",
                "CCDAMP" : "A",
                "CCDGAIN" : "1.0",
                "DATE-OBS" : "2002-03-19",
                "TIME-OBS" : "00:34:32",
                }) is None

    def test_rmap_todict(self):
        p = rmap.get_cached_mapping("hst.pmap")
        p.todict()

    def test_rmap_match_tjson(self):
        os.environ["CRDS_PATH"] = test_config.CRDS_TESTING_CACHE
        p = rmap.get_cached_mapping("jwst.pmap")
        p.tojson()

    def _get_rmap(self):
        log.set_verbose(55)
        return rmap.ReferenceMapping.from_string("""
header = {
    'derived_from' : 'hst_wfc3_darkfile_0379.rmap',
    'filekind' : 'DARKFILE',
    'instrument' : 'WFC3',
    'mapping' : 'REFERENCE',
    'name' : 'hst_wfc3_darkfile_0379.rmap',
    'observatory' : 'HST',
    'comment_parkeys' : ('BINAXIS1',),
    'parkey' : (('DETECTOR', 'CCDAMP', 'BINAXIS1', 'BINAXIS2', 'CCDGAIN', 'SAMP_SEQ', 'SUBTYPE'), ('DATE-OBS', 'TIME-OBS')),
    'parkey_relevance' : {
        'binaxis1' : '(DETECTOR == "UVIS")',
        'binaxis2' : '(DETECTOR == "UVIS")',
        'ccdgain' : '(DETECTOR == "IR")',
        'samp_seq' : '(DETECTOR == "IR")',
        'subtype' : '(DETECTOR == "IR")',
    },
    'reffile_format' : 'IMAGE',
    'reffile_required' : 'NONE',
    'reffile_switch' : 'DARKCORR',
    'rmap_relevance' : '(DARKCORR != "OMIT")',
    'sha1sum' : '16cfa985b83a7fb9db414dc8f339a95b9c03c5fa',
}

selector = Match({
    ('IR', 'ABCD', 'N/A', 'N/A', 1.0, 'MIF1200', 'FULLIMAG') : UseAfter({
        '2008-02-19 00:00:00' : 't3n16499i_drk.fits',
        '2008-02-20 00:00:00' : 't3n1649ai_drk.fits',
    }),
    ('UVIS', 'ABCD', 1, 1, 'N/A', 'N/A', 'N/A') : UseAfter({
        '2007-12-17 00:00:00' : 's9q1628ci_drk.fits',
        '2007-12-18 00:00:00' : 's9q1628ei_drk.fits',
        '2008-02-19 00:00:00' : 't3420175i_drk.fits',
        '2008-02-20 00:00:00' : 't3420177i_drk.fits',
    }),
})
""", ignore_checksum=True)

    def test_ref_to_dataset_ir(self):

        # rmap update time conversions don't map comment parkeys to N/A

        r = self._get_rmap()

        header = dict(
            DETECTOR="IR", CCDAMP="ABCD", BINAXIS1="1.0", BINAXIS2="2.0",
            CCDGAIN="1.0", SAMP_SEQ="MIF1200", SUBTYPE="FULLIMAG", DARKCORR="PERFORM")

        # transform for rmap updates
        dheader = utils.Struct(r.reference_to_dataset_header(header))

        self.assertEqual(dheader.DETECTOR, "IR")
        self.assertEqual(dheader.CCDAMP, "ABCD")
        self.assertEqual(dheader.BINAXIS1, "1.0")   # comment
        self.assertEqual(dheader.BINAXIS2, "N/A")   # non-comment
        self.assertEqual(dheader.CCDGAIN, "1.0")
        self.assertEqual(dheader.SAMP_SEQ, "MIF1200")
        self.assertEqual(dheader.SUBTYPE, "FULLIMAG")
        self.assertEqual(dheader.DARKCORR, "N/A")

    def test_ref_to_dataset_uvis(self):

        r = self._get_rmap()

        header = dict(
            DETECTOR="UVIS", CCDAMP="ABCD", BINAXIS1="1.0", BINAXIS2="2.0",
            CCDGAIN="1.0", SAMP_SEQ="MIF1200", SUBTYPE="FULLIMAG", DARKCORR="PERFORM")

        # transform for rmap updates
        dheader = utils.Struct(r.reference_to_dataset_header(header))

        self.assertEqual(dheader.DETECTOR, "UVIS")
        self.assertEqual(dheader.CCDAMP, "ABCD")
        self.assertEqual(dheader.BINAXIS1, "1.0")   # comment
        self.assertEqual(dheader.BINAXIS2, "2.0")   # non-comment
        self.assertEqual(dheader.CCDGAIN, "N/A")
        self.assertEqual(dheader.SAMP_SEQ, "N/A")
        self.assertEqual(dheader.SUBTYPE, "N/A")
        self.assertEqual(dheader.DARKCORR, "N/A")

    def test_na_parkeys_ir(self):

        r = self._get_rmap()

        # bestrefs time conversions still map comment parkeys to N/A

        header = dict(
            DETECTOR="IR", CCDAMP="ABCD", BINAXIS1="1.0", BINAXIS2="2.0",
            CCDGAIN="1.0", SAMP_SEQ="MIF1200", SUBTYPE="FULLIMAG", DARKCORR="PERFORM")

        # transform for bestrefs
        dheader = utils.Struct(r.map_irrelevant_parkeys_to_na(header))

        self.assertEqual(dheader.DETECTOR, "IR")
        self.assertEqual(dheader.CCDAMP, "ABCD")
        self.assertEqual(dheader.BINAXIS1, "N/A")   # comment
        self.assertEqual(dheader.BINAXIS2, "N/A")   # non-comment
        self.assertEqual(dheader.CCDGAIN, "1.0")
        self.assertEqual(dheader.SAMP_SEQ, "MIF1200")
        self.assertEqual(dheader.SUBTYPE, "FULLIMAG")
        self.assertEqual(dheader.DARKCORR, "PERFORM")

    def test_na_parkeys_uvis(self):

        r = self._get_rmap()

        header = dict(
            DETECTOR="UVIS", CCDAMP="ABCD", BINAXIS1="1.0", BINAXIS2="2.0",
            CCDGAIN="1.0", SAMP_SEQ="MIF1200", SUBTYPE="FULLIMAG", DARKCORR="PERFORM")

        # transform for bestrefs
        dheader = utils.Struct(r.map_irrelevant_parkeys_to_na(header))

        self.assertEqual(dheader.DETECTOR, "UVIS")
        self.assertEqual(dheader.CCDAMP, "ABCD")
        self.assertEqual(dheader.BINAXIS1, "N/A")   # comment
        self.assertEqual(dheader.BINAXIS2, "2.0")   # non-comment
        self.assertEqual(dheader.CCDGAIN, "N/A")
        self.assertEqual(dheader.SAMP_SEQ, "N/A")
        self.assertEqual(dheader.SUBTYPE, "N/A")
        self.assertEqual(dheader.DARKCORR, "PERFORM")

# ==================================================================================


def tst():
    """Run module tests,  for now just doctests only."""
    import unittest
    suite = unittest.TestLoader().loadTestsFromTestCase(TestRmap)
    unittest.TextTestRunner().run(suite)

    from crds.tests import test_rmap, tstmod
    return tstmod(test_rmap)

if __name__ == "__main__":
    print(tst())



'''
Created on Feb 16, 2017

@author: jmiller
'''
# import asdf

# ============================================================================

from crds.core import timestamp, utils, log

from .abstract import AbstractFile

# ============================================================================

class AsdfFile(AbstractFile):

    format = "ASDF"

    @utils.gc_collected
    def get_raw_header(self, needed_keys=(), **keys):
        """Return the flattened header associated with an ASDF file."""
        import asdf
        with asdf.open(self.filepath) as handle:
            header = self.to_simple_types(handle.tree)
            header["HISTORY"] = self.get_history(handle)
        return header

    def get_history(self, handle):
        """Given and ASDF file object `handle`, return the history collected into a
        single string.
        """
        history = "UNDEFINED"   # or BAD FORMAT
        with log.error_on_exception(
                "Failed reading ASDF history, see ASDF docs on adding history"):
            histall = []
            hist = handle.tree["history"]
            try:
                entries = handle.get_history_entries()
            except Exception:
                log.verbose_warning(
                    "Using inlined CRDS ASDF history entry reading interface.")
                entries = hist["entries"] if "entries" in hist else hist
            for entry in entries:
                time = timestamp.format_date(entry["time"]).split(".")[0]
                description = entry["description"]
                histall.append(time + " :: " + description)
            if histall:
                history = "\n".join(histall)
        return history

    def get_asdf_standard_version(self):
        """
        Return the ASDF Standard version associated with this file as a string,
        or `None` if the file is neither an ASDF file nor contains an embedded
        ASDF file.
        """
        import asdf
        with asdf.open(self.filepath) as handle:
            return str(handle.version)



"""This module manages the automatic generation of new context files based on
a list of new rmaps and a baseline context.
"""
import os.path
import sys
import shutil
import re
import glob

import crds
from crds.core import (rmap, utils, log, cmdline, config)
from . import refactor

# =============================================================================

# Code used by the website to determine affected

def get_update_map(old_pipeline, updated_rmaps):
    """Given the name of a pipeline context, `old_pipeline`, and a list
    of new rmap names, `updated_rmaps`,  return the mapping:

        { imap_name : [ updates_for_that_imap, ... ], ... }

    Updated rmaps can be rmap names or strings of the form:

        <instrument>_<filekind>_"n/a"

    e.g.  miri_dflat_n/a
    """
    pctx = crds.get_pickled_mapping(old_pipeline)   # reviewed
    updates = {}
    for update in sorted(updated_rmaps):
        if update.endswith(("_n/a","_N/A")):
            instrument, _filekind, na = update.split("_")
        else:  # should be an rmap name
            instrument, _filekind = utils.get_file_properties(pctx.observatory, update)
        imap_name = pctx.get_imap(instrument).filename
        if imap_name not in updates:
            updates[imap_name] = []
        assert update not in updates[imap_name], "Duplicate update for " + repr(update)
        updates[imap_name].append(update)
    return updates

# =============================================================================

# Code used by the website to generate new contexts.

def generate_new_contexts(old_pipeline, updates, new_names):
    """Generate new pipeline and instrument context files given:
    old_pipeline -- name of pipeline mapping
    updates --   { old_imap : [ new_rmaps ], ... }
    new_names --   { old_pmap : new_pmap, old_imaps : new_imaps }
    """
    new_names = dict(new_names)
    for imap_name in sorted(updates):
        hack_in_new_maps(imap_name, new_names[imap_name], updates[imap_name])
    new_pipeline = new_names.pop(old_pipeline)
    new_imaps = list(new_names.values())
    hack_in_new_maps(old_pipeline, new_pipeline, new_imaps)
    return [new_pipeline] + new_imaps

def hack_in_new_maps(old, new, updated_maps):
    """Given mapping named `old`,  create a modified copy named `new` which
    installs each map of `updated_maps` in place of it's predecessor.
    """
    copy_mapping(old, new)
    for mapping in sorted(updated_maps):
        key, replaced, replacement = insert_mapping(new, mapping)
        if replaced:
            log.info("Replaced", repr(replaced), "with", repr(replacement), "for", repr(key), "in", repr(old), "producing", repr(new))
        else:
            log.info("Added", repr(replacement), "for", repr(key), "in", repr(old), "producing", repr(new))

def insert_mapping(context, mapping):
    """Replace the filename in file `context` with the same generic name
    as `mapping` with `mapping`.  Re-write `context` in place.

    If mapping is of the form <instrument>_<type>_"n/a",  then it specifies
    that <type> of <instrument> should be set to "N/A".
    """
    # 'ACS' : 'hst_acs.imap',
    where = config.locate_mapping(context)
    # readonly caching is ok because this call is always made on a newly named
    # copy of the original rmap;  the only thing mutated is the uncached new mapping.
    loaded = rmap.asmapping(context, cache="readonly")
    if mapping.endswith(("_n/a", "_N/A")):
        instrument, key, special = mapping.split("_")
        replacement = special.upper()
    else:
        key = loaded.get_item_key(mapping)
        replacement = mapping
    key, replaced = loaded.set_item(key, os.path.basename(replacement))
    loaded.write(where)
    return key, replaced, replacement

def copy_mapping(old_map, new_map):
    """Make a copy of mapping `old_map` named `new_map`."""
    old_path = config.locate_mapping(old_map)
    new_path = config.locate_mapping(new_map)
    assert not os.path.exists(new_path), "New mapping file " + repr(new_map) + " already exists."
    shutil.copyfile(old_path, new_path)

# =============================================================================

# Code for making "fake/test" new contexts on the command line.

def new_context(old_pipeline, updated_rmaps):
    """Given a pipeline mapping name `old_pipeline`, and a list of the names
    of new rmaps, `updated_rmaps`, generate new imaps as needed and a single
    pmap which refers to them all.

    Returns  { old_name : fake_names }
    """
    updates = get_update_map(old_pipeline, updated_rmaps)
    new_names = generate_fake_names(old_pipeline, updates)
    generate_new_contexts(old_pipeline, updates, new_names)
    return new_names

def generate_fake_names(old_pipeline, updates):
    """Generate a map from old pipeline and instrument context names to new
    names for their updated replacements.   "Fake" names only work locally
    and may collide with CRDS server names...  and hence would not be
    submissible.
    """
    new_names = {}
    new_names[old_pipeline] = fake_name(old_pipeline)
    for old_imap in sorted(updates):
        new_names[old_imap] = fake_name(old_imap)
    return new_names

def fake_name(old_map):
    """Given and old mapping name, `old_map`, adjust the serial number to
    create a new mapping name of the same series.   This name is fake in the
    sense that it is local to a developer's machine.
    """
    if re.search(r"_\d+", old_map):
        map_glob = re.sub(r"_\d+(\..map)", r"_*\1", old_map)
        same_maps = sorted(glob.glob(config.locate_mapping(map_glob)))
        if same_maps:
            last_map = same_maps[-1]
            match = re.search(r"_(\d+)\..map", last_map)
            serial = int(match.group(1), 10) + 1
            new_map = re.sub(r"_(\d+)(\.[pir]map)", r"_%04d\2" % serial, old_map)
        else:
            new_map = old_map
    elif re.search(r"\w+[^\d]+\..map", old_map):
        # if no serial,  start off existing sequence as 0001
        parts = os.path.splitext(old_map)
        new_map = parts[0] + "_0001" + parts[1]
        new_map = fake_name(new_map)
    else:
        raise ValueError("Unrecognized mapping filename " + repr(old_map))
    if os.path.exists(config.locate_mapping(new_map)):
        # recurse until there's a free name,  or eventually fail.
        return fake_name(new_map)
    else:
        if not new_map.startswith("./"):
            new_map = "./" + os.path.basename(new_map)
        return new_map

def update_header_names(name_map):
    """Update the .name and .derived_from fields in mapping new_path.header
    to reflect derivation from old_path and name new_path.
    """
    for old_path, new_path in sorted(name_map.items()):
        old_base, new_base = os.path.basename(old_path), os.path.basename(new_path)
        refactor.update_derivation(new_path, old_base)
        log.info("Adjusting name", repr(new_base), "derived_from", repr(old_base),
                 "in", repr(new_path))
    return name_map # no change

# ============================================================================

class NewContextScript(cmdline.Script):
    """Defines the command line handler for crds newcontext."""

    description = """Based on `old_pmap`,  generate a new .pmap and .imaps as
needed in order to support `new_rmaps`.   Currently generated contexts have
fake names and are for local test purposes only,  not formal distribution.
"""
    def add_args(self):
        self.add_argument("old_pmap")
        self.add_argument("new_rmap", nargs="+", help="Names of new rmaps to insert into the new context.""")

    def main(self):
        name_map = new_context(self.args.old_pmap, self.args.new_rmap)
        update_header_names(name_map)
        return log.errors()

if __name__ == "__main__":
    sys.exit(NewContextScript()())



import os
import tempfile

# ==================================================================================
import numpy as np

from nose.tools import assert_raises, assert_true

# ==================================================================================

from crds.core import utils, log, exceptions
from crds.certify import reftypes
from crds import hst, jwst, roman

from crds.tests import test_config

# ==================================================================================


def reftypes_load_type_spec_spec():
    """
    >>> old_state = test_config.setup()
    >>> SPEC_FILE = os.path.join(os.path.abspath(hst.HERE), "specs", "acs_biasfile.spec")
    >>> spec = reftypes.TypeSpec.from_file(SPEC_FILE)
    >>> test_config.cleanup(old_state)
    """

def reftypes_load_type_spec_rmap():
    """
    >>> old_state = test_config.setup()
    >>> SPEC_FILE = os.path.join(os.path.abspath(hst.HERE), "specs", "cos_xwlkfile.rmap")
    >>> spec = reftypes.TypeSpec.from_file(SPEC_FILE)
    >>> test_config.cleanup(old_state)
    """

def reftypes_hst_load_raw_specs():
    """
    >>> old_state = test_config.setup()
    >>> SPECS = os.path.join(os.path.abspath(hst.HERE), "specs")
    >>> spec = reftypes.load_raw_specs(SPECS)
    >>> test_config.cleanup(old_state)
    """

def reftypes_hst_save_json_specs():
    """
    >>> old_state = test_config.setup()
    >>> SPECS = os.path.join(os.path.abspath(hst.HERE), "specs")
    >>> specs = reftypes.load_raw_specs(SPECS)
    >>> f = tempfile.NamedTemporaryFile(delete=False)
    >>> f.close()
    >>> reftypes.save_json_specs(specs, f.name)  # doctest: +ELLIPSIS
    CRDS - INFO -  Saved combined type specs to '...'
    >>> test_config.cleanup(old_state)
    """

def reftypes_jwst_load_raw_specs():
    """
    >>> old_state = test_config.setup()
    >>> SPECS = os.path.join(os.path.abspath(jwst.HERE), "specs")
    >>> spec = reftypes.load_raw_specs(SPECS)
    >>> test_config.cleanup(old_state)
    """

def reftypes_jwst_save_json_specs():
    """
    >>> old_state = test_config.setup()
    >>> SPECS = os.path.join(os.path.abspath(jwst.HERE), "specs")
    >>> specs = reftypes.load_raw_specs(SPECS)
    >>> f = tempfile.NamedTemporaryFile(delete=False)
    >>> f.close()
    >>> reftypes.save_json_specs(specs, f.name) # doctest: +ELLIPSIS
    CRDS - INFO -  Saved combined type specs to '...'
    >>> test_config.cleanup(old_state)
    """

def reftypes_roman_load_raw_specs():
    """
    >>> old_state = test_config.setup()
    >>> SPECS = os.path.join(os.path.abspath(roman.HERE), "specs")
    >>> spec = reftypes.load_raw_specs(SPECS)
    >>> test_config.cleanup(old_state)
    """

def reftypes_roman_save_json_specs():
    """
    >>> old_state = test_config.setup()
    >>> SPECS = os.path.join(os.path.abspath(roman.HERE), "specs")
    >>> specs = reftypes.load_raw_specs(SPECS)
    >>> f = tempfile.NamedTemporaryFile(delete=False)
    >>> f.close()
    >>> reftypes.save_json_specs(specs, f.name) # doctest: +ELLIPSIS
    CRDS - INFO -  Saved combined type specs to '...'
    >>> test_config.cleanup(old_state)
    """

def reftypes_hst_reference_name_to_tpn_infos():
    """
    >>> old_state = test_config.setup()
    >>> types = reftypes.get_types_object("hst")
    >>> infos = types.reference_name_to_tpninfos("data/s7g1700gl_dead.fits")
    >>> print(log.PP(infos))
    [('DESCRIP', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('DETECTOR', 'HEADER', 'CHARACTER', 'REQUIRED', values=('FUV', 'NUV')),
     ('FILETYPE', 'HEADER', 'CHARACTER', 'REQUIRED', values=('DEADTIME REFERENCE TABLE',)),
     ('INSTRUME', 'HEADER', 'CHARACTER', 'REQUIRED', values=('COS',)),
     ('PEDIGREE', 'HEADER', 'CHARACTER', 'REQUIRED', values=('&PEDIGREE',)),
     ('SEGMENT', 'COLUMN', 'CHARACTER', 'REQUIRED', values=('FUVA', 'FUVB', 'ANY')),
     ('USEAFTER', 'HEADER', 'CHARACTER', 'REQUIRED', values=('&SYBDATE',)),
     ('VCALCOS', 'HEADER', 'CHARACTER', 'REQUIRED', values=())]
    >>> test_config.cleanup(old_state)
    """

def reftypes_jwst_reference_name_to_tpn_infos():    # doctest: +ELLIPSIS
    """
    >>> old_state = test_config.setup()
    >>> types = reftypes.get_types_object("jwst")
    >>> infos = types.reference_name_to_tpninfos("data/jwst_miri_flat_slitlessprism.fits")
    >>> print(log.PP(infos))
    [('DETECTOR', 'HEADER', 'CHARACTER', 'OPTIONAL', values=()),
     ('DQ', 'ARRAY_FORMAT', 'EXPRESSION', condition='((array_exists(SCI_ARRAY))and(array_exists(DQ_ARRAY)))', expression='(SCI_ARRAY.SHAPE[-2:]==DQ_ARRAY.SHAPE[-2:])'),
     ('DQ', 'ARRAY_FORMAT', 'EXPRESSION', 'OPTIONAL', expression="(warn_only(has_type(DQ_ARRAY,'INT')))"),
     ('DQ', 'ARRAY_FORMAT', 'EXPRESSION', 'WARN', expression='(is_image(DQ_ARRAY))'),
     ('DQ_DEF', 'ARRAY_FORMAT', 'EXPRESSION', 'OPTIONAL', expression="(has_column_type(DQ_DEF_ARRAY,'BIT','INT'))"),
     ('DQ_DEF', 'ARRAY_FORMAT', 'EXPRESSION', 'OPTIONAL', expression="(has_column_type(DQ_DEF_ARRAY,'DESCRIPTION','STRING'))"),
     ('DQ_DEF', 'ARRAY_FORMAT', 'EXPRESSION', 'OPTIONAL', expression="(has_column_type(DQ_DEF_ARRAY,'NAME','STRING'))"),
     ('DQ_DEF', 'ARRAY_FORMAT', 'EXPRESSION', 'OPTIONAL', expression="(has_column_type(DQ_DEF_ARRAY,'VALUE','INT'))"),
     ('DQ_DEF', 'ARRAY_FORMAT', 'EXPRESSION', 'OPTIONAL', expression="(has_columns(DQ_DEF_ARRAY,['BIT','VALUE','NAME','DESCRIPTION']))"),
     ('DQ_DEF', 'ARRAY_FORMAT', 'EXPRESSION', 'OPTIONAL', expression='(is_table(DQ_DEF_ARRAY))'),
     ('ERR', 'ARRAY_FORMAT', 'EXPRESSION', 'REQUIRED', expression='(SCI_ARRAY.SHAPE[-2:]==ERR_ARRAY.SHAPE[-2:])'),
     ('ERR', 'ARRAY_FORMAT', 'EXPRESSION', 'REQUIRED', expression="(has_type(ERR_ARRAY,'FLOAT'))"),
     ('ERR', 'ARRAY_FORMAT', 'EXPRESSION', 'REQUIRED', expression='(is_image(ERR_ARRAY))'),
     ('EXP_TYPE', 'HEADER', 'CHARACTER', 'OPTIONAL', values=()),
     ('FULLFRAME_XSIZE', 'EXPRESSION', 'EXPRESSION', 'IF_FULL_FRAME', expression='(META_SUBARRAY_XSIZE==1032)'),
     ('FULLFRAME_XSTART', 'EXPRESSION', 'EXPRESSION', 'IF_FULL_FRAME', expression='(META_SUBARRAY_XSTART==1)'),
     ('FULLFRAME_YSIZE', 'EXPRESSION', 'EXPRESSION', 'IF_FULL_FRAME', expression='(META_SUBARRAY_YSIZE==1024)'),
     ('FULLFRAME_YSTART', 'EXPRESSION', 'EXPRESSION', 'IF_FULL_FRAME', expression='(META_SUBARRAY_YSTART==1)'),
     ('META.AUTHOR', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('META.DESCRIPTION', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('META.EXPOSURE.READPATT', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('ACQ1', 'ACQ2', 'BRIGHT1', 'BRIGHT2', 'DEEP2', 'DEEP8', 'FAST', 'FASTGRPAVG', 'FASTGRPAVG8', 'FASTGRPAVG16', 'FASTGRPAVG32', 'FASTGRPAVG64', 'FASTR1', 'FASTR100FGS', 'FGS60', 'FGS8370', 'FGS840', 'FGSRAPID', 'FINEGUIDE', 'ID', 'MEDIUM2', 'MEDIUM8', 'NIS', 'NISRAPID', 'NRS', 'NRSIRS2', 'NRSN16R4', 'NRSN32R8', 'NRSN8R2', 'NRSRAPID', 'NRSIRS2RAPID', 'NRSRAPIDD1', 'NRSRAPIDD2', 'NRSRAPIDD6', 'NRSSLOW', 'RAPID', 'SHALLOW2', 'SHALLOW4', 'SLOW', 'SLOWR1', 'TRACK', 'ANY', 'N/A')),
     ('META.EXPOSURE.TYPE', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('MIR_4QPM', 'MIR_CORONCAL', 'MIR_DARK', 'MIR_DARKALL', 'MIR_DARKIMG', 'MIR_DARKMRS', 'MIR_FLATALL', 'MIR_FLATIMAGE', 'MIR_FLAT-IMAGE', 'MIR_FLATIMAGE-EXT', 'MIR_FLATMRS', 'MIR_FLAT-MRS', 'MIR_FLATMRS-EXT', 'MIR_IMAGE', 'MIR_LRS-FIXEDSLIT', 'MIR_LRS-SLITLESS', 'MIR_LYOT', 'MIR_MRS', 'MIR_TACONFIRM', 'MIR_TACQ', 'ANY', 'N/A')),
     ('META.HISTORY', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('META.INSTRUMENT.BAND', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('SHORT', 'MEDIUM', 'LONG', 'SHORT-MEDIUM', 'SHORT-LONG', 'MEDIUM-SHORT', 'MEDIUM-LONG', 'LONG-SHORT', 'LONG-MEDIUM', 'MULTIPLE', 'ANY', 'N/A')),
     ('META.INSTRUMENT.CHANNEL', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('1', '2', '3', '4', '12', '34', '1234', '123', '234', 'ANY', 'N/A')),
     ('META.INSTRUMENT.CORONAGRAPH', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('4QPM', '4QPM_1065', '4QPM_1140', '4QPM_1550', 'LYOT', 'LYOT_2300', 'MASKA210R', 'MASKASWB', 'MASKA335R', 'MASKA430R', 'MASKALWB', 'NONE', 'ANY', 'N/A')),
     ('META.INSTRUMENT.DETECTOR', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('MIRIFULONG', 'MIRIFUSHORT', 'MIRIMAGE', 'ANY', 'N/A')),
     ('META.INSTRUMENT.FILTER', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('CLEAR', 'F070LP', 'F070W', 'F090W', 'F1000W', 'F100LP', 'F1065C', 'F110W', 'F1130W', 'F1140C', 'F115W', 'F1280W', 'F140M', 'F140X', 'F1500W', 'F150W', 'F150W2', 'F1550C', 'F170LP', 'F1800W', 'F182M', 'F187N', 'F200W', 'F2100W', 'F210M', 'F212N', 'F2300C', 'F250M', 'F2550W', 'F2550WR', 'F277W', 'F290LP', 'F300M', 'F322W2', 'F335M', 'F356W', 'F360M', 'F380M', 'F410M', 'F430M', 'F444W', 'F460M', 'F480M', 'F560W', 'F770W', 'FLENS', 'FND', 'GR150C', 'GR150R', 'OPAQUE', 'P750L', 'WLP4', 'MULTIPLE', 'ANY', 'N/A')),
     ('META.INSTRUMENT.GRATING', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('G140M', 'G235M', 'G395M', 'G140H', 'G235H', 'G395H', 'PRISM', 'MIRROR', 'UNKNOWN', 'MULTIPLE', 'N/A', 'ANY')),
     ('META.INSTRUMENT.NAME', 'HEADER', 'CHARACTER', 'REQUIRED', values=('MIRI',)),
     ('META.INSTRUMENT.PUPIL', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('CLEAR', 'CLEARP', 'F090W', 'F115W', 'F140M', 'F150W', 'F158M', 'F162M', 'F164N', 'F200W', 'F323N', 'F405N', 'F466N', 'F470N', 'FLAT', 'GDHS0', 'GDHS60', 'GR700XD', 'GRISMC', 'GRISMR', 'GRISMV2', 'GRISMV3', 'MASKBAR', 'MASKIPR', 'MASKRND', 'NRM', 'PINHOLES', 'WLM8', 'WLP8', 'ANY', 'N/A')),
     ('META.MODEL_TYPE', 'HEADER', 'CHARACTER', condition="(warning(not(META_REFTYPE.startswith('pars-'))))", values=()),
     ('META.PEDIGREE', 'HEADER', 'CHARACTER', 'REQUIRED', values=('&JWSTPEDIGREE',)),
     ('META.REFTYPE', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('META.SUBARRAY.FASTAXIS', 'HEADER', 'INTEGER', 'OPTIONAL', values=('1', '-1', '2', '-2')),
     ('META.SUBARRAY.FASTAXIS', 'HEADER', 'INTEGER', 'REQUIRED', values=()),
     ('META.SUBARRAY.NAME', 'HEADER', 'CHARACTER', 'OPTIONAL', values=('BRIGHTSKY', 'MASK1065', 'MASK1140', 'MASK1550', 'MASKLYOT', 'SLITLESSPRISM', 'SUB128', 'SUB256', 'SUB64', 'FULL', 'GENERIC', 'ANY', 'N/A')),
     ('META.SUBARRAY.NAME', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('META.SUBARRAY.SLOWAXIS', 'HEADER', 'INTEGER', 'OPTIONAL', values=('1', '-1', '2', '-2')),
     ('META.SUBARRAY.SLOWAXIS', 'HEADER', 'INTEGER', 'REQUIRED', values=()),
     ('META.SUBARRAY.XSIZE', 'HEADER', 'INTEGER', 'OPTIONAL', values=('1:1032',)),
     ('META.SUBARRAY.XSIZE', 'HEADER', 'INTEGER', 'REQUIRED', values=()),
     ('META.SUBARRAY.XSTART', 'HEADER', 'INTEGER', 'OPTIONAL', values=('1:1032',)),
     ('META.SUBARRAY.XSTART', 'HEADER', 'INTEGER', 'REQUIRED', values=()),
     ('META.SUBARRAY.YSIZE', 'HEADER', 'INTEGER', 'OPTIONAL', values=('1:1024',)),
     ('META.SUBARRAY.YSIZE', 'HEADER', 'INTEGER', 'REQUIRED', values=()),
     ('META.SUBARRAY.YSTART', 'HEADER', 'INTEGER', 'OPTIONAL', values=('1:1024',)),
     ('META.SUBARRAY.YSTART', 'HEADER', 'INTEGER', 'REQUIRED', values=()),
     ('META.TELESCOPE', 'HEADER', 'CHARACTER', 'REQUIRED', values=('JWST',)),
     ('META.USEAFTER', 'HEADER', 'CHARACTER', 'REQUIRED', values=('&JWSTDATE',)),
     ('MIRIFULONG_AXIS', 'EXPRESSION', 'EXPRESSION', condition="(DETECTOR=='MIRIFULONG')", expression='((FASTAXIS==1)and(SLOWAXIS==2))'),
     ('MIRIFUSHORT_AXIS', 'EXPRESSION', 'EXPRESSION', condition="(DETECTOR=='MIRIFUSHORT')", expression='((FASTAXIS==1)and(SLOWAXIS==2))'),
     ('MIRIMAGE_AXIS', 'EXPRESSION', 'EXPRESSION', condition="(DETECTOR=='MIRIMAGE')", expression='((FASTAXIS==1)and(SLOWAXIS==2))'),
     ('SCI', 'ARRAY_FORMAT', 'EXPRESSION', 'IF_FULL_FRAME', expression='(SCI_ARRAY.SHAPE[-2:]==(1024,1032))'),
     ('SCI', 'ARRAY_FORMAT', 'EXPRESSION', 'REQUIRED', expression="(has_type(SCI_ARRAY,'FLOAT'))"),
     ('SCI', 'ARRAY_FORMAT', 'EXPRESSION', 'REQUIRED', expression='(is_image(SCI_ARRAY))'),
     ('SCI', 'ARRAY_FORMAT', 'EXPRESSION', 'IF_SUBARRAY', expression='(1<=META_SUBARRAY_XSTART+SCI_ARRAY.SHAPE[-1]-1<=1032)'),
     ('SCI', 'ARRAY_FORMAT', 'EXPRESSION', 'IF_SUBARRAY', expression='(1<=META_SUBARRAY_YSTART+SCI_ARRAY.SHAPE[-2]-1<=1024)'),
     ('SCI', 'ARRAY_FORMAT', 'EXPRESSION', 'IF_SUBARRAY', expression='(SCI_ARRAY.SHAPE[-2:]==(META_SUBARRAY_YSIZE,META_SUBARRAY_XSIZE))'),
     ('SUBARRAY_INBOUNDS_X', 'EXPRESSION', 'EXPRESSION', 'ANY_SUBARRAY', expression='(1<=META_SUBARRAY_XSTART+META_SUBARRAY_XSIZE-1<=1032)'),
     ('SUBARRAY_INBOUNDS_Y', 'EXPRESSION', 'EXPRESSION', 'ANY_SUBARRAY', expression='(1<=META_SUBARRAY_YSTART+META_SUBARRAY_YSIZE-1<=1024)'),
     ('SUBARRAY_XSIZE', 'EXPRESSION', 'EXPRESSION', 'IF_SUBARRAY', expression='(1<=META_SUBARRAY_XSIZE<=1032)'),
     ('SUBARRAY_XSTART', 'EXPRESSION', 'EXPRESSION', 'IF_SUBARRAY', expression='(1<=META_SUBARRAY_XSTART<=1032)'),
     ('SUBARRAY_YSIZE', 'EXPRESSION', 'EXPRESSION', 'IF_SUBARRAY', expression='(1<=META_SUBARRAY_YSIZE<=1024)'),
     ('SUBARRAY_YSTART', 'EXPRESSION', 'EXPRESSION', 'IF_SUBARRAY', expression='(1<=META_SUBARRAY_YSTART<=1024)')]
    >>> test_config.cleanup(old_state)
    """

def reftypes_roman_reference_name_to_tpn_infos():
    """
    >>> old_state = test_config.setup() # doctest: +ELLIPSIS
    >>> types = reftypes.get_types_object("roman")
    >>> infos = types.reference_name_to_tpninfos("roman_wfi_flat.asdf")
    >>> print(log.PP(infos))
    [('ROMAN.META.AUTHOR', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('ROMAN.META.DESCRIPTION', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('ROMAN.META.EXPOSURE.TYPE', 'HEADER', 'CHARACTER', 'OPTIONAL', values=(...)),
     ('ROMAN.META.INSTRUMENT.DETECTOR', 'HEADER', 'CHARACTER', 'OPTIONAL', values=(...)),
     ('ROMAN.META.INSTRUMENT.DETECTOR', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('ROMAN.META.INSTRUMENT.NAME', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('ROMAN.META.INSTRUMENT.NAME', 'HEADER', 'CHARACTER', 'REQUIRED', values=('WFI',)),
     ('ROMAN.META.INSTRUMENT.OPTICAL_ELEMENT', 'HEADER', 'CHARACTER', 'OPTIONAL', values=(...)),
     ('ROMAN.META.INSTRUMENT.OPTICAL_ELEMENT', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('ROMAN.META.ORIGIN', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('ROMAN.META.PEDIGREE', 'HEADER', 'CHARACTER', 'REQUIRED', values=('&JWSTPEDIGREE',)),
     ('ROMAN.META.REFTYPE', 'HEADER', 'CHARACTER', 'REQUIRED', values=()),
     ('ROMAN.META.TELESCOPE', 'HEADER', 'CHARACTER', 'REQUIRED', values=('ROMAN',)),
     ('ROMAN.META.USEAFTER', 'HEADER', 'CHARACTER', 'REQUIRED', values=('&JWSTDATE',))]
    >>> test_config.cleanup(old_state)
    """

def reftypes_hst_get_filekinds():
    """
    >>> old_state = test_config.setup()
    >>> types = reftypes.get_types_object("hst")
    >>> types.get_filekinds("nicmos")
    ['backtab', 'darkfile', 'flatfile', 'idctab', 'illmfile', 'maskfile', 'nlinfile', 'noisfile', 'pedsbtab', 'phottab', 'pmodfile', 'pmskfile', 'rnlcortb', 'saacntab', 'saadfile', 'tdffile', 'tempfile', 'zprattab']
    >>> test_config.cleanup(old_state)
    """

def reftypes_jwst_get_filekinds():
    """
    >>> old_state = test_config.setup()
    >>> types = reftypes.get_types_object("jwst")
    >>> types.get_filekinds("niriss")
    ['abvegaoffset', 'all', 'amplifier', 'apcorr', 'area', 'dark', 'distortion', 'drizpars', 'extract1d', 'flat', 'gain', 'ipc', 'linearity', 'mask', 'pars-darkpipeline', 'pars-detector1pipeline', 'pars-image2pipeline', 'pars-outlierdetectionstep', 'pars-sourcecatalogstep', 'pars-spec2pipeline', 'pars-tweakregstep', 'pathloss', 'persat', 'photom', 'readnoise', 'regions', 'saturation', 'speckernel', 'specprofile', 'spectrace', 'specwcs', 'superbias', 'throughput', 'trapdensity', 'trappars', 'wavelengthrange', 'wavemap', 'wcsregions', 'wfssbkg']
    >>> test_config.cleanup(old_state)
    """

def reftypes_roman_get_filekinds():
    """
    >>> old_state = test_config.setup()
    >>> types = reftypes.get_types_object("roman")
    >>> {'all', 'flat'}.issubset(types.get_filekinds("wfi"))
    True
    >>> test_config.cleanup(old_state)
    """

def reftypes_reference_name_to_tpn_text():
    """
    >>> old_state = test_config.setup()
    >>> types = reftypes.get_types_object("hst")
    >>> print(types.reference_name_to_tpn_text("data/s7g1700gl_dead.fits"))
    From TPN: cos_dead.tpn
    ----------------------
    INSTRUME            H        C         R    COS
    FILETYPE            H        C         R    "DEADTIME REFERENCE TABLE"
    DETECTOR            H        C         R    FUV,NUV
    VCALCOS             H        C         R
    USEAFTER            H        C         R    &SYBDATE
    PEDIGREE            H        C         R    &PEDIGREE
    DESCRIP             H        C         R
    SEGMENT             C        C         R    FUVA,FUVB,ANY
    <BLANKLINE>

    >>> test_config.cleanup(old_state)
    """

def reftypes_reference_name_to_ld_tpn_text():
    """
    >>> old_state = test_config.setup()
    >>> types = reftypes.get_types_object("hst")
    >>> print(types.reference_name_to_tpn_text("data/s7g1700gl_dead.fits"))
    From TPN: cos_dead.tpn
    ----------------------
    INSTRUME            H        C         R    COS
    FILETYPE            H        C         R    "DEADTIME REFERENCE TABLE"
    DETECTOR            H        C         R    FUV,NUV
    VCALCOS             H        C         R
    USEAFTER            H        C         R    &SYBDATE
    PEDIGREE            H        C         R    &PEDIGREE
    DESCRIP             H        C         R
    SEGMENT             C        C         R    FUVA,FUVB,ANY
    <BLANKLINE>
    >>> test_config.cleanup(old_state)
    """


def reftypes_get_row_keys_by_instrument():
    """
    >>> old_state = test_config.setup()
    >>> types = reftypes.get_types_object("hst")
    >>> types.get_row_keys_by_instrument("cos")
    ['aperture', 'cenwave', 'date', 'fpoffset', 'opt_elem', 'segment']
    >>> test_config.cleanup(old_state)
    """

# ==================================================================================

class TestReftypes(test_config.CRDSTestCase):

    def setUp(self, *args, **keys):
        super(TestReftypes, self).setUp(*args, **keys)
        self._old_debug = log.set_exception_trap(False)

    def tearDown(self, *args, **keys):
        super(TestReftypes, self).tearDown(*args, **keys)
        log.set_exception_trap(self._old_debug)

    # ------------------------------------------------------------------------------

    def test_validator_bad_presence(self):
        #         tinfo = certify.generic_tpn.TpnInfo('DETECTOR','H','C','Q', ('WFC','HRC','SBC'))
        #         assert_raises(ValueError, certify.validators.validator, tinfo)
        pass

# ==================================================================================

def main():
    """Run module tests,  for now just doctests only."""
    import unittest

    suite = unittest.TestLoader().loadTestsFromTestCase(TestReftypes)
    unittest.TextTestRunner().run(suite)

    from crds.tests import test_reftypes, tstmod
    return tstmod(test_reftypes)

if __name__ == "__main__":
    print(main())
